from __future__ import annotations

import tarfile

from dissect.hypervisor.util import vmtar


def main() -> None:
    # We just want to run the main function of the tarfile module, but with our VisorTarFile and is_tarfile functions
    type(tarfile.main)(
        tarfile.main.__code__,
        tarfile.main.__globals__
        | {
            "TarFile": vmtar.VisorTarFile,
            "is_tarfile": vmtar.is_tarfile,
            "open": vmtar.open,
        },
    )()


if __name__ == "__main__":
    main()



# References:
# - /bin/crypto-util
# - /usr/lib/vmware/tpm/bin/keypersist
# - /lib/libvmlibs.so

from __future__ import annotations

import hashlib
import io
from base64 import b64decode
from typing import BinaryIO, NamedTuple
from urllib.parse import unquote
from uuid import UUID

try:
    import _pystandalone

    HAS_PYSTANDALONE = True
except ImportError:
    HAS_PYSTANDALONE = False

try:
    from Crypto.Cipher import AES

    HAS_PYCRYPTODOME = True
except ImportError:
    HAS_PYCRYPTODOME = False

from dissect.cstruct import cstruct
from dissect.util.stream import RangeStream

c_def = """
struct EnvelopeFileHeader {
    char    magic[21];
    char    _pad[483];
    uint32  size;
    uint32  version;
};

struct DataTransformAeadFooter {
    char    magic[23];
    char    _pad[9];
    char    data[4056];
    uint32  size;
    uint32  version;
};

struct DataTransformCryptoFooter {
    char    magic[25];
    char    _pad[479];
    uint32  padding;
    uint32  version;
};

enum AttributeType : uint8 {
    Invalid = 0x0,
    // These are guesses based on size and used attributes
    UInt8 = 0x1,
    UInt16 = 0x2,
    UInt32 = 0x3,
    UInt64 = 0x4,
    Int8 = 0x5,
    Int16 = 0x6,
    Int32 = 0x7,
    Int64 = 0x8,
    Float = 0x9,
    Double = 0xA,
    // These are certain
    String = 0xB,
    Bytes = 0xC
};
"""
c_envelope = cstruct().load(c_def)

FILE_HEADER_MAGIC = b"DataTransformEnvelope"
FOOTER_AEAD_MAGIC = b"DataTransformAeadFooter"
FOOTER_CRYPTO_MAGIC = b"DataTransformCryptoFooter"

PBKDF2_SALT = b"This is obfuscation, not encryption. If you want encryption, use TPM."

ENVELOPE_BLOCK_SIZE = 4096
ENVELOPE_ATTRIBUTE_TYPE_MAP = {
    c_envelope.AttributeType.Invalid: None,
    c_envelope.AttributeType.UInt8: c_envelope.uint8,
    c_envelope.AttributeType.UInt16: c_envelope.uint16,
    c_envelope.AttributeType.UInt32: c_envelope.uint32,
    c_envelope.AttributeType.UInt64: c_envelope.uint64,
    c_envelope.AttributeType.Int8: c_envelope.int8,
    c_envelope.AttributeType.Int16: c_envelope.int16,
    c_envelope.AttributeType.Int32: c_envelope.int32,
    c_envelope.AttributeType.Int64: c_envelope.int64,
    c_envelope.AttributeType.Float: c_envelope.float,
    c_envelope.AttributeType.Double: c_envelope.double,
    c_envelope.AttributeType.String: None,
    c_envelope.AttributeType.Bytes: None,
}

DECRYPT_CHUNK_SIZE = 1024 * 1024 * 4


class EnvelopeAttribute(NamedTuple):
    type: int
    flag: int
    value: bytes


class Envelope:
    """Implements an encryption envelope as used within ESXi."""

    def __init__(self, fh: BinaryIO, verify: bool = True):
        self.fh = fh
        self.verify = verify

        header_buf = io.BytesIO(self.fh.read(ENVELOPE_BLOCK_SIZE))
        self.header = c_envelope.EnvelopeFileHeader(header_buf)

        if self.header.magic != FILE_HEADER_MAGIC:
            raise ValueError("Invalid envelope file")

        if self.header.version != 2:
            raise ValueError("Unsupported envelope type")

        self.version = self.header.version
        self.attributes = _read_envelope_attributes(header_buf)
        for req in ("vmware.keyInfo", "vmware.cipherName", "vmware.keyHash"):
            if req not in self.attributes:
                raise ValueError(f"Missing required {req} attribute")

        self.key_info = self.attributes["vmware.keyInfo"].value
        self.cipher_name = self.attributes["vmware.cipherName"].value
        self.key_hash = self.attributes["vmware.keyHash"].value
        self.iv = self.attributes.get("vmware.iv", EnvelopeAttribute(None, None, None)).value
        self.digest = None

        if self.cipher_name == "AES-256-GCM":
            self.fh.seek(-ENVELOPE_BLOCK_SIZE, io.SEEK_END)
            aead_footer = c_envelope.DataTransformAeadFooter(self.fh)
            if aead_footer.version != 1:
                raise ValueError("Invalid AEAD footer")
            self.digest = aead_footer.data[: aead_footer.size]
        else:
            raise NotImplementedError(f"Unsupported cipher: {self.cipher_name}")

        self.fh.seek(0, io.SEEK_END)
        size = self.fh.tell()
        self.size = size - (2 * ENVELOPE_BLOCK_SIZE)
        self.data = RangeStream(self.fh, ENVELOPE_BLOCK_SIZE, self.size)

    def decrypt(self, key: bytes, aad: bytes | None = None) -> bytes:
        """Decrypt the data in this envelope.

        Arguments:
            key: decryption key to use
            aad: optional associated data to include for AEAD ciphers
        """
        if not HAS_PYSTANDALONE and not HAS_PYCRYPTODOME:
            raise RuntimeError("No crypto module available")

        key_hash = hashlib.sha256(self.cipher_name.encode() + key).digest()
        if key_hash != self.key_hash:
            raise ValueError("Key hash doesn't match")

        is_aead = False
        if self.cipher_name == "AES-256-GCM":
            is_aead = True

            if not self.iv:
                raise ValueError("Missing IV")

            if HAS_PYSTANDALONE:
                cipher = _pystandalone.aes_256_gcm(key, self.iv)
            else:
                cipher = AES.new(key, AES.MODE_GCM, nonce=self.iv)
        else:
            raise NotImplementedError(f"Unsupported cipher: {self.cipher_name}")

        if is_aead:
            # The file header is included in the AAD, as well as any optional variable data
            cipher.update(_pack_envelope_header(self))
            if aad:
                cipher.update(aad)

        self.data.seek(0)
        offset = 0
        decrypted = bytearray(self.size)
        while True:
            chunk = self.data.read(DECRYPT_CHUNK_SIZE)
            if not chunk:
                break

            chunk_size = len(chunk)
            decrypted[offset : offset + chunk_size] = cipher.decrypt(chunk)
            offset += chunk_size

        footer = c_envelope.DataTransformCryptoFooter(bytes(decrypted[-512:]))
        decrypted = decrypted[: -4096 - footer.padding]

        if self.verify:
            cipher.verify(self.digest)

        return bytes(decrypted)


class KeyStore:
    """Implements a file based keystore as used in ESXi."""

    def __init__(self, store: dict[str, str]):
        self.store = store

        self.mode = self.store.get("mode", None)
        if not self.mode:
            raise ValueError("Keystore has no mode")

        self._id = None
        self._key = None

        if self.mode == "NONE":
            data = self.store["ConfigEncData"]
            obj = {}
            for opt in data.split(":"):
                name, _, value = opt.partition("=")
                obj[name.strip()] = unquote(value.strip())

            self._id = str(UUID(bytes=b64decode(obj["keyId"])))

            data1 = b64decode(obj["data1"])
            data2 = b64decode(obj["data2"])

            self._key = hashlib.pbkdf2_hmac("sha256", data1 + PBKDF2_SALT, data2, 100000)
        else:
            raise NotImplementedError("Only NONE is implemented")

    @property
    def id(self) -> str:
        return self._id

    @property
    def key(self) -> bytes:
        return self._key

    @classmethod
    def from_text(cls, text: str) -> KeyStore:
        """Parse a key store from a string.

        Arguments:
            text: string to parse a key store from
        """
        store = {}

        for line in text.split("\n"):
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            name, _, value = line.partition("=")

            name = name.strip()
            value = value.strip(' "')

            if name.startswith("."):
                store[name] = value
            else:
                node = store
                parts = name.strip().split(".")
                for part in parts[:-1]:
                    if part not in node:
                        node[part] = {}
                    node = node[part]

                node[parts[-1]] = value.strip(' "')

        return cls(store)


def _read_envelope_attributes(buf: BinaryIO) -> dict[str, EnvelopeAttribute]:
    attributes = {}
    while True:
        try:
            attr_type = c_envelope.AttributeType(buf)
            if attr_type == c_envelope.AttributeType.Invalid:
                break

            flag = c_envelope.uint8(buf)
            buf.read(2)

            name = c_envelope.char[None](buf).decode()
            if attr_type in ENVELOPE_ATTRIBUTE_TYPE_MAP:
                if attr_type == c_envelope.AttributeType.String:
                    value = c_envelope.char[None](buf).decode()
                elif attr_type == c_envelope.AttributeType.Bytes:
                    value = buf.read(c_envelope.uint64(buf))
                else:
                    value = ENVELOPE_ATTRIBUTE_TYPE_MAP[attr_type](buf)
            else:
                raise NotImplementedError(f"Unknown attribute type: {attr_type}")

            attributes[name] = EnvelopeAttribute(attr_type, flag, value)
        except EOFError:
            break

    return attributes


def _pack_envelope_header(envelope: Envelope, block_size: int = ENVELOPE_BLOCK_SIZE) -> bytes:
    stream = io.BytesIO()

    stream.write(b"\x00" * 512)

    _pack_attributes(stream, envelope.attributes)

    size = stream.tell()
    remainder = size % block_size
    if remainder:
        stream.write(b"\x00" * (block_size - remainder))
    size = stream.tell()

    stream.seek(0)
    c_envelope.EnvelopeFileHeader(
        magic=b"DataTransformEnvelope",
        size=size - len(c_envelope.EnvelopeFileHeader),
        version=envelope.version,
    ).write(stream)

    return stream.getvalue()


def _pack_attributes(stream: BinaryIO, attributes: dict[str, EnvelopeAttribute]) -> None:
    for name, attribute in attributes.items():
        c_envelope.AttributeType.write(stream, attribute.type)
        c_envelope.uint8.write(stream, attribute.flag)
        c_envelope.uint16.write(stream, 0)
        c_envelope.char[None].write(stream, name.encode())

        if attribute.type == c_envelope.AttributeType.String:
            c_envelope.char[None].write(stream, attribute.value.encode())
        elif attribute.type == c_envelope.AttributeType.Bytes:
            c_envelope.uint64.write(stream, len(attribute.value))
            stream.write(attribute.value)
        else:
            ENVELOPE_ATTRIBUTE_TYPE_MAP[attribute.type].write(stream, attribute.value)

    # Null terminate attributes
    stream.write(b"\x00" * 4)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_vdi(__cs__.cstruct):
    VDI_IMAGE_COMMENT_SIZE: Literal[256] = ...
    class VDI_IMAGE_TYPE(__cs__.Enum):
        NORMAL = ...
        FIXED = ...
        UNDO = ...
        DIFF = ...

    class VDI_IMAGE_FLAGS(__cs__.Flag):
        ZERO_EXPAND = ...

    class VDIDISKGEOMETRY(__cs__.Structure):
        Cylinders: _c_vdi.uint32
        Heads: _c_vdi.uint32
        Sectors: _c_vdi.uint32
        Sector: _c_vdi.uint32
        @overload
        def __init__(
            self,
            Cylinders: _c_vdi.uint32 | None = ...,
            Heads: _c_vdi.uint32 | None = ...,
            Sectors: _c_vdi.uint32 | None = ...,
            Sector: _c_vdi.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    PVDIDISKGEOMETRY: TypeAlias = __cs__.Pointer[_c_vdi.VDIDISKGEOMETRY]
    class VDIPREHEADER(__cs__.Structure):
        szFileInfo: __cs__.CharArray
        u32Signature: _c_vdi.uint32
        u32Version: _c_vdi.uint32
        @overload
        def __init__(
            self,
            szFileInfo: __cs__.CharArray | None = ...,
            u32Signature: _c_vdi.uint32 | None = ...,
            u32Version: _c_vdi.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    PVDIPREHEADER: TypeAlias = __cs__.Pointer[_c_vdi.VDIPREHEADER]
    class VDIHEADER1PLUS(__cs__.Structure):
        cbHeader: _c_vdi.uint32
        u32Type: _c_vdi.VDI_IMAGE_TYPE
        fFlags: _c_vdi.VDI_IMAGE_FLAGS
        szComment: __cs__.CharArray
        offBlocks: _c_vdi.uint32
        offData: _c_vdi.uint32
        LegacyGeometry: _c_vdi.VDIDISKGEOMETRY
        u32Dummy: _c_vdi.uint32
        cbDisk: _c_vdi.uint64
        cbBlock: _c_vdi.uint32
        cbBlockExtra: _c_vdi.uint32
        cBlocks: _c_vdi.uint32
        cBlocksAllocated: _c_vdi.uint32
        uuidCreate: __cs__.CharArray
        uuidModify: __cs__.CharArray
        uuidLinkage: __cs__.CharArray
        uuidParentModify: __cs__.CharArray
        Geometry: _c_vdi.VDIDISKGEOMETRY
        @overload
        def __init__(
            self,
            cbHeader: _c_vdi.uint32 | None = ...,
            u32Type: _c_vdi.VDI_IMAGE_TYPE | None = ...,
            fFlags: _c_vdi.VDI_IMAGE_FLAGS | None = ...,
            szComment: __cs__.CharArray | None = ...,
            offBlocks: _c_vdi.uint32 | None = ...,
            offData: _c_vdi.uint32 | None = ...,
            LegacyGeometry: _c_vdi.VDIDISKGEOMETRY | None = ...,
            u32Dummy: _c_vdi.uint32 | None = ...,
            cbDisk: _c_vdi.uint64 | None = ...,
            cbBlock: _c_vdi.uint32 | None = ...,
            cbBlockExtra: _c_vdi.uint32 | None = ...,
            cBlocks: _c_vdi.uint32 | None = ...,
            cBlocksAllocated: _c_vdi.uint32 | None = ...,
            uuidCreate: __cs__.CharArray | None = ...,
            uuidModify: __cs__.CharArray | None = ...,
            uuidLinkage: __cs__.CharArray | None = ...,
            uuidParentModify: __cs__.CharArray | None = ...,
            Geometry: _c_vdi.VDIDISKGEOMETRY | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    PVDIHEADER1PLUS: TypeAlias = __cs__.Pointer[_c_vdi.VDIHEADER1PLUS]

# Technically `c_vdi` is an instance of `_c_vdi`, but then we can't use it in type hints
c_vdi: TypeAlias = _c_vdi

VDI_IMAGE_SIGNATURE: int
VDI_IMAGE_BLOCK_FREE: int
VDI_IMAGE_BLOCK_ZERO: int



from __future__ import annotations

from dissect.cstruct import cstruct

asif_def = """
#define ASIF_HEADER_SIGNATURE       0x73686477  // 'shdw'
#define ASIF_META_HEADER_SIGNATURE  0x6D657461  // 'meta'

struct asif_header {
    uint32  header_signature;
    uint32  header_version;
    uint32  header_size;
    uint32  header_flags;
    uint64  directory_offsets[2];
    char    guid[16];
    uint64  sector_count;
    uint64  max_sector_count;
    uint32  chunk_size;
    uint16  block_size;
    uint16  total_segments;
    uint64  metadata_chunk;
    char    unk_50[16];
    uint32  read_only_flags;
    uint32  metadata_flags;
    uint32  metadata_read_only_flags;
};

struct asif_meta_header {
    uint32  header_signature;
    uint32  header_version;
    uint32  header_size;
    uint64  data_size;
    uint64  unk_14;
};
"""

c_asif = cstruct(endian=">").load(asif_def)



from __future__ import annotations

import argparse
import sys
from pathlib import Path

from dissect.hypervisor.util.envelope import Envelope, KeyStore


def main() -> int:
    parser = argparse.ArgumentParser(description="ESXi envelope file decrypter")
    parser.add_argument("envelope", type=Path, help="envelope file")
    parser.add_argument("-ks", "--keystore", type=Path, required=True, help="keystore file")
    parser.add_argument("-o", "--output", type=Path, required=True, help="output file")
    args = parser.parse_args()

    if not args.envelope.exists() or not args.keystore.exists():
        parser.exit("Need both envelope and keystore file")

    with args.envelope.open("rb") as fh:
        envelope = Envelope(fh)
        keystore = KeyStore.from_text(args.keystore.read_text())

        with args.output.open("wb") as fhout:
            fhout.write(envelope.decrypt(keystore.key))

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream
from dissect.util.xmemoryview import xmemoryview

from dissect.hypervisor.disk.c_vdi import VDI_IMAGE_BLOCK_FREE, VDI_IMAGE_BLOCK_ZERO, VDI_IMAGE_SIGNATURE, c_vdi
from dissect.hypervisor.exceptions import Error

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self


class VDI:
    """VirtualBox Virtual Disk Image (VDI) implementation.

    Use :meth:`open` to get a stream for reading from the VDI file. The stream will handle reading
    from the parent disk if necessary (and provided).

    If provided with a file-like object, the caller is responsible for closing it.
    When provided with a path, the VDI class will manage the file handle.

    If providing a parent file-like object, the caller is responsible for the lifecycle of that object.

    Args:
        fh: File-like object or path of the VDI file.
        parent: Optional file-like object for the parent disk (for differencing disks).
    """

    def __init__(self, fh: BinaryIO | Path, parent: BinaryIO | None = None):
        if isinstance(fh, Path):
            self.path = fh
            self.fh = self.path.open("rb")
        else:
            self.path = None
            self.fh = fh

        self.parent = parent

        self.fh.seek(0)
        self.preheader = c_vdi.VDIPREHEADER(self.fh)
        if self.preheader.u32Signature != VDI_IMAGE_SIGNATURE:
            raise Error(
                f"Invalid VDI signature, expected {VDI_IMAGE_SIGNATURE:#08X}, got {self.preheader.u32Signature:#08X}"
            )

        self.header = c_vdi.VDIHEADER1PLUS(self.fh)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self.close()

    @property
    def type(self) -> c_vdi.VDI_IMAGE_TYPE:
        """The type of the VDI file."""
        return self.header.u32Type

    @property
    def flags(self) -> c_vdi.VDI_IMAGE_FLAGS:
        """The flags of the VDI file."""
        return self.header.fFlags

    @property
    def size(self) -> int:
        """The size of the virtual disk."""
        return self.header.cbDisk

    @property
    def block_size(self) -> int:
        """The size of each block in the VDI file."""
        return self.header.cbBlock

    @property
    def data_offset(self) -> int:
        """The offset to the data blocks."""
        return self.header.offData

    @property
    def blocks_offset(self) -> int:
        """The offset to the block allocation table."""
        return self.header.offBlocks

    @property
    def number_of_blocks(self) -> int:
        """The number of blocks in the VDI file."""
        return self.header.cBlocks

    def open(self) -> VDIStream:
        """Open a stream to read from the VDI file."""
        return VDIStream(self)

    def close(self) -> None:
        """Close the VDI file handle."""
        if self.path is not None:
            self.fh.close()


class VDIStream(AlignedStream):
    def __init__(self, vdi: VDI):
        self.vdi = vdi
        self.block_size = vdi.block_size

        self.fh = self.vdi.fh
        self.fh.seek(self.vdi.blocks_offset)
        self.map = xmemoryview(self.fh.read(4 * self.vdi.number_of_blocks), "<i")
        super().__init__(vdi.size, align=self.block_size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        block_idx, offset_in_block = divmod(offset, self.block_size)
        while length > 0:
            read_len = min(length, max(length, self.block_size - offset_in_block))

            block = self.map[block_idx]
            if block == VDI_IMAGE_BLOCK_FREE:
                if self.vdi.parent is not None:
                    self.vdi.parent.seek(offset)
                    result.append(self.vdi.parent.read(read_len))
                else:
                    result.append(b"\x00" * read_len)
            elif block == VDI_IMAGE_BLOCK_ZERO:
                result.append(b"\x00" * read_len)
            else:
                self.fh.seek(self.vdi.data_offset + (block * self.block_size) + offset_in_block)
                result.append(self.fh.read(read_len))

            offset += read_len
            length -= read_len
            block_idx += 1

        return b"".join(result)



# References:
# - /bin/vmtar

from __future__ import annotations

import struct
import tarfile
from io import BytesIO
from typing import BinaryIO, Final


class VisorTarInfo(tarfile.TarInfo):
    """Implements TarInfo for use with Visor Tar files (vmtar).

    The main difference is that file data is located at the end of the tar file, on
    an offset specified in the header.
    """

    is_visor: bool
    offset_data: int | None
    textPgs: int | None
    fixUpPgs: int | None

    @classmethod
    def _init_visor_attrs(cls, obj: VisorTarInfo, buf: bytes) -> None:
        """Initialize visor-specific attributes from the raw header buffer."""
        obj.is_visor = buf[257:264] == b"visor  "
        if obj.is_visor:
            obj.offset_data = struct.unpack("<I", buf[496:500])[0]
            obj.textPgs = struct.unpack("<I", buf[504:508])[0]
            obj.fixUpPgs = struct.unpack("<I", buf[508:512])[0]
        else:
            obj.offset_data = None
            obj.textPgs = None
            obj.fixUpPgs = None

    @classmethod
    def frombuf(cls, buf: bytes, encoding: str, errors: str) -> VisorTarInfo:
        obj = super().frombuf(buf, encoding, errors)
        cls._init_visor_attrs(obj, buf)
        return obj

    @classmethod
    def _frombuf(cls, buf: bytes, encoding: str, errors: str, **kwargs) -> VisorTarInfo:
        # Python 3.13.13+ refactored tarfile to call _frombuf directly instead of
        # frombuf in fromtarfile and _proc_gnulong/_proc_pax. We must override _frombuf
        # to ensure visor-specific attributes are initialized for all code paths.
        obj = super()._frombuf(buf, encoding, errors, **kwargs)
        cls._init_visor_attrs(obj, buf)
        return obj

    def _proc_member(self, tarfile: tarfile.TarFile) -> VisorTarInfo | tarfile.TarInfo:
        if self.is_visor and self.offset_data:
            # Don't advance the offset with the filesize
            tarfile.offset = tarfile.fileobj.tell()

            # Patch the TarInfo object with saved global
            # header information.
            self._apply_pax_info(tarfile.pax_headers, tarfile.encoding, tarfile.errors)

            return self

        return super()._proc_member(tarfile)


class VisorTarFile(tarfile.TarFile):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, tarinfo=VisorTarInfo)

    @classmethod
    def visoropen(cls, name: str, mode: str = "r", fileobj: BinaryIO | None = None, **kwargs) -> VisorTarFile:
        """Open a visor tar file for reading. Supports gzip and lzma compression."""
        if mode not in ("r",):
            raise tarfile.TarError("visor currently only supports read mode")

        try:
            from gzip import GzipFile
        except ImportError:
            raise tarfile.CompressionError("gzip module is not available") from None

        try:
            from lzma import LZMAError, LZMAFile
        except ImportError:
            raise tarfile.CompressionError("lzma module is not available") from None

        compressed = False

        try:
            t = cls.taropen(name, mode, fileobj, **kwargs)
        except Exception:
            if fileobj is not None:
                fileobj.seek(0)
            try:
                fileobj = GzipFile(name, mode + "b", fileobj=fileobj)
            except OSError as e:
                if fileobj is not None and mode == "r":
                    raise tarfile.ReadError("not a visor file") from e
                raise

            try:
                t = cls.taropen(name, mode, fileobj, **kwargs)
            except Exception:
                fileobj.seek(0)
                fileobj = LZMAFile(fileobj or name, mode)  # noqa: SIM115

                try:
                    t = cls.taropen(name, mode, fileobj, **kwargs)
                except (LZMAError, EOFError, OSError) as e:
                    fileobj.close()
                    if mode == "r":
                        raise tarfile.ReadError("not a visor file") from e
                    raise
                except:
                    fileobj.close()
                    raise

            compressed = True

        # If we get here, we have a valid visor tar file
        if fileobj is not None:
            fileobj.seek(0)
            if compressed:
                # Read the decompressed data into a BytesIO to make random reads faster
                fileobj = BytesIO(fileobj.read())

        t = cls.taropen(name, mode, fileobj, **kwargs)

        t._extfileobj = False
        return t

    # Only allow opening visor tar files
    OPEN_METH: Final[dict[str, str]] = {"visor": "visoropen"}


open = VisorTarFile.open

is_tarfile = type(tarfile.is_tarfile)(tarfile.is_tarfile.__code__, tarfile.is_tarfile.__globals__ | {"open": open})



from __future__ import annotations

from uuid import UUID

from dissect.cstruct import cstruct

vhdx_def = """
#define PAYLOAD_BLOCK_NOT_PRESENT           0
#define PAYLOAD_BLOCK_UNDEFINED             1
#define PAYLOAD_BLOCK_ZERO                  2
#define PAYLOAD_BLOCK_UNMAPPED              3
#define PAYLOAD_BLOCK_FULLY_PRESENT         6
#define PAYLOAD_BLOCK_PARTIALLY_PRESENT     7

#define SB_BLOCK_NOT_PRESENT                0
#define SB_BLOCK_PRESENT                    6

struct file_identifier {
    char    signature[8];
    char    creator[512];
};

struct header {
    char    signature[4];
    uint32  checksum;
    uint64  sequence_number;
    char    file_write_guid[16];
    char    data_write_guid[16];
    char    log_guid[16];
    uint16  log_version;
    uint16  version;
    uint32  log_length;
    uint64  log_offset;
    char    reserved[4096];
};

struct region_table_header {
    char    signature[4];
    uint32  checksum;
    uint32  entry_count;
    char    reserved[4];
};

struct region_table_entry {
    char    guid[16];
    uint64  file_offset;
    uint32  length;
    uint32  required;
};

struct bat_entry {
    uint64  state:3;
    uint64  reserved:17;
    uint64  file_offset_mb:44;
};

struct metadata_table_header {
    char    signature[8];
    char    reserved[2];
    uint16  entry_count;
    char    reserved2[20];
};

struct metadata_table_entry {
    char    item_id[16];
    uint32  offset;
    uint32  length;
    uint32  is_user:1;
    uint32  is_virtual_disk:1;
    uint32  is_required:1;
    uint32  reserved:29;
    uint32  reserved2:2;
};

struct file_parameters {
    uint32  block_size;
    uint32  leave_block_allocated:1;
    uint32  has_parent:1;
    uint32  reserved:30;
};

struct virtual_disk_id {
    char    virtual_disk_id[16];            // typedef of an array is still broken
};

typedef uint64      virtual_disk_size;
typedef uint32      logical_sector_size;
typedef uint32      physical_sector_size;

struct parent_locator_header {
    char    locator_type[16];
    uint16  reserved;
    uint16  key_value_count;
};

struct parent_locator_entry {
    uint32  key_offset;
    uint32  value_offset;
    uint16  key_length;
    uint16  value_length;
};
"""

c_vhdx = cstruct().load(vhdx_def)

ALIGNMENT = 64 * 1024
MB = 1024 * 1024

BAT_REGION_GUID = UUID("2DC27766-F623-4200-9D64-115E9BFD4A08")
FILE_PARAMETERS_GUID = UUID("CAA16737-FA36-4D43-B3B6-33F0AA44E76B")
LOGICAL_SECTOR_SIZE_GUID = UUID("8141BF1D-A96F-4709-BA47-F233A8FAAB5F")
METADATA_REGION_GUID = UUID("8B7CA206-4790-4B9A-B8FE-575F050F886E")
PARENT_LOCATOR_GUID = UUID("A8D35F2D-B30B-454D-ABF7-D3D84834AB0C")
PHYSICAL_SECTOR_SIZE_GUID = UUID("CDA348C7-445D-4471-9CC9-E9885251C556")
VIRTUAL_DISK_ID_GUID = UUID("BECA12AB-B2E6-4523-93EF-C309E000C746")
VIRTUAL_DISK_SIZE_GUID = UUID("2FA54224-CD1B-4876-B211-5DBED83BF4B8")

VHDX_PARENT_LOCATOR_GUID = UUID("B04AEFB7-D19E-4A81-B789-25B8E9445913")



from __future__ import annotations

import io
import struct
from functools import lru_cache
from typing import BinaryIO

from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_vhd import SECTOR_SIZE, c_vhd


def read_footer(fh: BinaryIO) -> c_vhd.footer:
    fh.seek(-512, io.SEEK_END)
    footer = c_vhd.footer(fh)
    if not footer.features & 0x00000002:
        # Versions previous to Microsoft Virtual PC 2004 can have a 511 byte footer
        fh.seek(-511, io.SEEK_END)
        footer = c_vhd.footer(fh)
    return footer


class VHD(AlignedStream):
    # Note: split VHD files are currently unsupported.
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        footer = read_footer(fh)
        if footer.data_offset == 0xFFFFFFFFFFFFFFFF:
            # Fixed size, data starts at 0
            self.disk = FixedDisk(fh, footer)
        else:
            self.disk = DynamicDisk(fh, footer)

        super().__init__(self.disk.size)

    def _read(self, offset: int, length: int) -> bytes:
        sector = offset // SECTOR_SIZE
        count = (length + SECTOR_SIZE - 1) // SECTOR_SIZE

        return self.disk.read_sectors(sector, count)


class Disk:
    def __init__(self, fh: BinaryIO, footer: c_vhd.footer | None = None):
        self.fh = fh
        self.footer = footer if footer else read_footer(fh)
        self.size = self.footer.current_size

    def read_sectors(self, sector: int, count: int) -> bytes:
        raise NotImplementedError


class FixedDisk(Disk):
    def read_sectors(self, sector: int, count: int) -> bytes:
        self.fh.seek(sector * SECTOR_SIZE)
        return self.fh.read(count * SECTOR_SIZE)


class DynamicDisk(Disk):
    def __init__(self, fh: BinaryIO, footer: c_vhd.footer | None = None):
        super().__init__(fh, footer)
        fh.seek(self.footer.data_offset)
        self.header = c_vhd.dynamic_header(fh)
        self.bat = BlockAllocationTable(fh, self.header.table_offset, self.header.max_table_entries)

        self._sectors_per_block = self.header.block_size // SECTOR_SIZE
        # Sector bitmaps are padded to SECTOR_SIZE boundaries
        # Save bitmap size in sectors
        self._sector_bitmap_size = ((self._sectors_per_block // 8) + SECTOR_SIZE - 1) // SECTOR_SIZE

    def read_sectors(self, sector: int, count: int) -> bytes:
        result = []
        while count > 0:
            block, offset = divmod(sector, self._sectors_per_block)
            block_remaining = self._sectors_per_block - offset

            read_count = min(count, block_remaining)

            sector_offset = self.bat[block]
            if sector_offset:
                self.fh.seek((sector_offset + self._sector_bitmap_size + offset) * SECTOR_SIZE)
                result.append(self.fh.read(read_count * SECTOR_SIZE))
            else:
                # Sparse
                result.append((b"\x00" * SECTOR_SIZE) * read_count)

            sector += read_count
            count -= read_count

        return b"".join(result)


class BlockAllocationTable:
    """Implementation of the BAT.

    Entries are uint32 sector offsets to blocks in the file.
    """

    ENTRY = struct.Struct(">I")

    def __init__(self, fh: BinaryIO, offset: int, max_entries: int):
        self.fh = fh
        self.offset = offset
        self.max_entries = max_entries

        self.get = lru_cache(4096)(self.get)

    def get(self, block: int) -> int | None:
        # This could be improved by caching the entire BAT (or chunks if too large)
        if block + 1 > self.max_entries:
            raise ValueError(f"Invalid block {block} (max block is {self.max_entries - 1})")

        self.fh.seek(self.offset + block * 4)
        sector_offset = self.ENTRY.unpack(self.fh.read(4))[0]
        if sector_offset == 0xFFFFFFFF:
            sector_offset = None
        return sector_offset

    def __getitem__(self, block: int) -> int | None:
        return self.get(block)



# Reference:
# - crypto-util
# - libvmlibs.so
from __future__ import annotations

import base64
import hashlib
import hmac
import re
from urllib.parse import unquote

try:
    import _pystandalone  # type: ignore

    HAS_PYSTANDALONE = True
except ImportError:
    HAS_PYSTANDALONE = False

try:
    from Crypto.Cipher import AES

    HAS_PYCRYPTODOME = True
except ImportError:
    HAS_PYCRYPTODOME = False


CIPHER_KEY_SIZES = {
    "AES-256": 32,
    "AES-192": 24,
    "AES-128": 16,
}

HMAC_MAP = {
    "HMAC-SHA-1": ("sha1", 20),
    "HMAC-SHA-1-128": ("sha1", 16),
    "HMAC-SHA-256": ("sha256", 32),
}

PASS2KEY_MAP = {
    "PBKDF2-HMAC-SHA-1": "sha1",
    "PBKDF2-HMAC-SHA-256": "sha256",
}


class VMX:
    def __init__(self, attr: dict[str, str]):
        self.attr = attr

    @classmethod
    def parse(cls, string: str) -> VMX:
        """Parse a VMX dictionary from a string."""
        return cls(_parse_dictionary(string))

    @property
    def encrypted(self) -> bool:
        """Return whether this VMX is encrypted.

        Encrypted VMXs will have both a ``encryption.keySafe`` and ``encryption.data`` value.
        The ``encryption.keySafe`` is a string encoded ``KeySafe``, which is made up of key locators.

        For example:

        .. code-block:: none

            vmware:key/list/(pair/(phrase/phrase_id/phrase_content,hmac,data),pair/(.../...,...,...))

        A ``KeySafe`` must be a list of ``Pairs``. Each ``Pair`` has a wrapped key, an HMAC type and encrypted data.
        It's implementation specific how to unwrap a key. E.g. a phrase is just PBKDF2. The unwrapped key
        can be used to unlock the encrypted ``Pair`` data. This will contain the final encryption key to decrypt
        the data in ``encryption.data``.

        So, in summary, to unseal a ``KeySafe``:

        - Parse ``KeySafe``
        - Iterate pairs
        - Unlock ``Pair``
        - Unwrap key (e.g. ``Phrase``)
        - Decrypt ``Pair`` data
        - Parse dictionary

        The terms for unwrapping, unlocking and unsealing are taken from VMware.
        """
        return "encryption.keysafe" in self.attr

    def unlock_with_phrase(self, passphrase: str) -> None:
        """Unlock this VMX in-place with a passphrase if it's encrypted.

        This will load the ``KeySafe`` from the current dictionary and attempt to recover the encryption key
        from it using the given passphrase. This key is used to decrypt the encrypted VMX data.

        The dictionary is updated in-place with the encrypted VMX data.
        """
        if not self.encrypted:
            raise TypeError("VMX is not encrypted")

        safe = KeySafe.from_text(self.attr["encryption.keysafe"])
        key, mac = safe.unseal_with_phrase(passphrase)

        encrypted = base64.b64decode(self.attr["encryption.data"])
        decrypted = _decrypt_hmac(key, encrypted, mac)
        self.attr.update(**_parse_dictionary(decrypted.decode()))

    def disks(self) -> list[str]:
        """Return a list of paths to disk files."""
        dev_classes = ("scsi", "sata", "ide", "nvme")
        devices = {}

        for vm_setting, value in self.attr.items():
            for dev_class in dev_classes:
                if vm_setting.startswith(dev_class):
                    # Properties for disk devices are formatted as
                    # <dev_class><bus_id>:<disk_id>.<dev_property>
                    #
                    # We use <bus_id>:<disk_id> as a unique identifier for
                    # disks to store the properties and their values.
                    # Properties for the bus device are stored with the unique
                    # <bus_id> key.
                    device, dev_property = vm_setting.split(".", 1)
                    dev_id = device.lstrip(dev_class)

                    dev_ids = devices.setdefault(dev_class, {})
                    dev_properties = dev_ids.setdefault(dev_id, {})
                    dev_properties[dev_property] = value
                    break

        disk_files = []
        for dev_properties_by_dev_id in devices.values():
            for dev_properties in dev_properties_by_dev_id.values():
                filename = dev_properties.get("filename")

                if filename:
                    dev_type = dev_properties.get("devicetype")

                    if not dev_type or "disk" in dev_type.lower():
                        disk_files.append(filename)

        return sorted(disk_files)


def _parse_dictionary(string: str) -> dict[str, str]:
    """Parse a VMX dictionary."""
    dictionary = {}

    for line in string.split("\n"):
        line = line.strip()

        if not line or line.startswith("#"):
            continue

        key, _, value = line.partition("=")
        # Some keys are technically case sensitive, but there are implementations
        # that have different casing for the same key
        dictionary[key.strip().lower()] = value.strip(' "')

    return dictionary


class KeySafe:
    def __init__(self, locators: list[Pair]):
        self.locators = locators

    def unseal_with_phrase(self, passphrase: str) -> bytes:
        """Unseal this ``KeySafe`` with a passphrase and return the decrypted key."""
        for locator in self.locators:
            if not locator.has_phrase():
                continue

            try:
                data = locator.unlock_with_phrase(passphrase)
                # Decrypted output is a crypto dict
                # type=key:cipher=AES-256:key=base64value'
                crypto_dict = _parse_crypto_dict(data.decode())
                return base64.b64decode(crypto_dict["key"]), locator.mac
            except ValueError:
                pass

        raise ValueError("No compatible locator")

    @classmethod
    def from_text(cls, text: str) -> KeySafe:
        """Parse a ``KeySafe`` from a string."""
        # Key safes are a list of key locators. It's a key locator string with a specific prefix
        identifier, _, remainder = text.partition("/")
        if identifier != "vmware:key":
            raise ValueError("Invalid KeySafe string, wrong identifier")

        # First part must be a list of pairs
        locators = _parse_key_locator(remainder)
        if not isinstance(locators, list) and not all(isinstance(member, Pair) for member in locators):
            raise ValueError("Invalid KeySafe string, not a list of pairs")

        return KeySafe(locators)


class Pair:
    def __init__(self, wrapped_key: Phrase, mac: str, data: bytes):
        self.wrapped_key = wrapped_key
        self.mac = mac
        self.data = data

    def __repr__(self) -> str:
        return f"<Pair wrapped_key={self.wrapped_key} mac={self.mac}>"

    def has_phrase(self) -> bool:
        """Return whether this ``Pair`` is a ``Phrase`` pair."""
        return isinstance(self.wrapped_key, Phrase)

    def _unlock(self, key: bytes) -> bytes:
        """Decrypt the data in this ``Pair``."""
        return _decrypt_hmac(key, self.data, self.mac)

    def unlock(self, *args, **kwargs) -> bytes:
        """Helper method to unlock this ``Pair`` for various wrapped keys.

        Currently only supports `Phrase`.
        """
        if self.has_phrase():
            return self.unlock_with_phrase(*args, **kwargs)

        raise TypeError(f"Unable to unlock {self.wrapped_key}")

    def unlock_with_phrase(self, passphrase: str) -> bytes:
        """Unlock this ``Pair`` with a passphrase and return the decrypted data."""
        if not self.has_phrase():
            raise TypeError("Pair doesn't have a phrase protected key")

        key = self.wrapped_key.unwrap(passphrase)
        return self._unlock(key)


class Phrase:
    def __init__(self, id: str, pass2key: str, cipher: str, rounds: int, salt: bytes):
        self.id = id
        self.pass2key = pass2key
        self.cipher = cipher
        self.rounds = rounds
        self.salt = salt

    def __repr__(self) -> str:
        return f"<Phrase id={self.id} pass2key={self.pass2key} cipher={self.cipher} rounds={self.rounds}>"

    def unwrap(self, passphrase: str) -> bytes:
        """Unwrap/generate the encryption key for a given passphrase.

        VMware calls this unwrapping, but really it's a KDF with the properties of this ``Phrase``.
        """
        return hashlib.pbkdf2_hmac(
            PASS2KEY_MAP[self.pass2key],
            passphrase.encode(),
            self.salt,
            self.rounds,
            CIPHER_KEY_SIZES[self.cipher],
        )


def _parse_key_locator(locator_string: str) -> Pair | Phrase | list[Pair | Phrase]:
    """Parse a key locator from a string.

    Key locators are string formatted data structures with a forward slash (``/``) separator. Each component is
    prefixed with a type, followed by that types' specific data. Values between separators are URL encoded.

    Interally called ``KeyLocator``.
    """
    identifier, _, remainder = locator_string.partition("/")

    if identifier == "list":
        # Comma separated list in between braces
        # list/(member,member)
        return [_parse_key_locator(member) for member in _split_list(remainder)]

    if identifier == "pair":
        # Comma separated tuple with 3 members
        # pair/(key data,mac type,encrypted data)
        members = _split_list(remainder)
        return Pair(
            _parse_key_locator(members[0]),
            unquote(members[1]),
            base64.b64decode(unquote(members[2])),
        )

    if identifier == "phrase":
        # Serialized crypto dict, prefixed with an identifier
        # phrase/encoded id/encoded dict
        phrase_id, _, phrase_data = remainder.partition("/")
        crypto_dict = _parse_crypto_dict(unquote(phrase_data))
        return Phrase(
            unquote(phrase_id),
            crypto_dict["pass2key"],
            crypto_dict["cipher"],
            int(crypto_dict["rounds"]),
            base64.b64decode(crypto_dict["salt"]),
        )

    # rawkey, ldap, script, role, fqid
    raise NotImplementedError(f"Not implemented keysafe identifier: {identifier}")


def _split_list(value: str) -> list[str]:
    """Parse a key locator list from a string.

    Lists are wrapped by braces and separated by comma. They can contain nested lists/pairs,
    so we need to separate at the correct nest level.
    """
    if not (match := re.match(r"\((.+)\)", value)):
        raise ValueError("Invalid list string")

    contents = match.group(1)

    buf = ""
    members = []
    level = 0
    for char in contents:
        if char == "(":
            level += 1
        elif char == ")":
            level -= 1
        elif char == "," and level == 0:
            members.append(buf)
            buf = ""
            continue

        buf += char

    if buf:
        members.append(buf)

    return members


def _parse_crypto_dict(dict_string: str) -> dict[str, str]:
    """Parse a crypto dict from a string.

    Crypto dicts are encoded as ``key=encoded_value:key=encoded_value``.

    Internally called ``CryptoDict``.
    """
    crypto_dict = {}
    for part in dict_string.split(":"):
        key, _, value = part.partition("=")
        crypto_dict[key] = unquote(value)
    return crypto_dict


def _decrypt_hmac(key: bytes, data: bytes, digest: str) -> bytes:
    """Decrypt and validate ciphertext.

    First 16 bytes of the ciphertext are the IV and the last N bytes are the HMAC digest.
    The cleartext is padded using PKCS#7.
    """
    digest, digest_size = HMAC_MAP[digest]

    iv, encrypted, mac = data[:16], data[16:-digest_size], data[-digest_size:]
    cipher = _create_cipher(key, iv)

    decrypted = cipher.decrypt(encrypted)
    if decrypted[-1] <= 16:
        # PKCS#7 padding
        decrypted = decrypted[: -decrypted[-1]]

    # We don't do any secret crypto so we don't care about the warning in the docs about timing attacks
    if hmac.digest(key, decrypted, digest) != mac:
        raise ValueError("Invalid HMAC, wrong key?")

    return decrypted


def _create_cipher(key: bytes, iv: bytes) -> AES.CbcMode:
    """Create a cipher object.

    Dynamic based on the available crypto module.
    """
    if HAS_PYSTANDALONE:
        if len(key) == 32:
            cipher = "aes-256-cbc"
        elif len(key) == 24:
            cipher = "aes-192-cbc"
        elif len(key) == 16:
            cipher = "aes-128-cbc"
        else:
            raise ValueError(f"Invalid key size: {len(key)}")

        return _pystandalone.cipher(cipher, key, iv)

    if HAS_PYCRYPTODOME:
        return AES.new(key, AES.MODE_CBC, iv=iv)

    raise RuntimeError("No crypto module available")



from __future__ import annotations

from typing import TYPE_CHECKING, TextIO

from defusedxml import ElementTree

if TYPE_CHECKING:
    from collections.abc import Iterator
    from xml.etree.ElementTree import Element


class PVS:
    """Parallels VM settings file.

    Args:
        fh: The file-like object to a PVS file.
    """

    def __init__(self, fh: TextIO):
        self._xml: Element = ElementTree.fromstring(fh.read())

    def disks(self) -> Iterator[str]:
        """Yield the disk file names."""
        for hdd_elem in self._xml.iterfind(".//Hdd"):
            system_name = hdd_elem.find("SystemName")
            if system_name is not None:
                yield system_name.text



from __future__ import annotations

import plistlib
from functools import cached_property, lru_cache
from typing import Any, BinaryIO
from uuid import UUID

from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_asif import c_asif
from dissect.hypervisor.exceptions import InvalidSignature


class ASIF:
    """Apple Sparse Image Format (ASIF) disk image.

    ASIF disk images are a virtual disk format introduced in macOS Tahoe. They can be used in Apple's Virtualization
    framework, as well as through Disk Utility.

    An ASIF file is pretty straight forward. There's a small header which, among some other details, contains two
    directory offsets. Each directory contains a list of tables, which in turn contain a list of data entries. Each data
    entry points to a chunk of data in the ASIF file. The chunk size is defined in the header and is typically 1 MiB.
    The chunk size is always a multiple of the block size, which is also defined in the header (typically 512 bytes).
    Each directory has a version number, and the directory with the highest version number is the active directory. This
    allows for atomic updates of the directory/table data.

    The maximum virtual disk size seems to be just under 4 PiB, with a small portion at the end reserved for metadata.
    The actual size of the virtual disk is defined in the header, as well as the maximum size the disk can grow to.

    The offset to the metadata block is typically ``(4 PiB - 1 chunk)``, meaning it's within the reserved area.
    The metadata block contains a small header and a plist. The plist should contain an ``internal metadata`` and
    ``user metadata`` dictionary. Besides a "stable uuid", it's unclear what the metadata is used for or how to set it.

    Args:
        fh: File-like object containing the ASIF image.

    Resources:
        - Reversing ``diskimagescontroller``
        - https://developer.apple.com/documentation/virtualization/vzdiskimagestoragedeviceattachment/
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self.header = c_asif.asif_header(fh)
        if self.header.header_signature != c_asif.ASIF_HEADER_SIGNATURE:
            raise InvalidSignature(
                f"Not a valid ASIF image (expected {c_asif.ASIF_HEADER_SIGNATURE:#x}, "
                f"got {self.header.header_signature:#x})"
            )

        self.guid = UUID(bytes=self.header.guid)
        self.block_size = self.header.block_size
        self.chunk_size = self.header.chunk_size
        self.size = self.header.sector_count * self.block_size
        self.max_size = self.header.max_sector_count * self.block_size

        # The following math is taken from the assembly with some creative variable naming
        # It's possible that some of this can be simplified or the names improved
        self._blocks_per_chunk = self.chunk_size // self.block_size

        # This check doesn't really make sense, but keep it in for now
        reserved_size = 4 * self.chunk_size
        self._num_reserved_table_entries = (
            1 if reserved_size < self._blocks_per_chunk else reserved_size // self._blocks_per_chunk
        )

        self._max_table_entries = self.chunk_size >> 3
        self._num_table_entries = self._max_table_entries - (
            self._max_table_entries % (self._num_reserved_table_entries + 1)
        )
        self._num_reserved_directory_entries = (self._num_reserved_table_entries + self._num_table_entries) // (
            self._num_reserved_table_entries + 1
        )
        self._num_usable_entries = self._num_table_entries - self._num_reserved_directory_entries
        # This is the size in bytes of data covered by a single table
        self._size_per_table = self._num_usable_entries * self.chunk_size

        max_size = self.block_size * self.header.max_sector_count
        self._num_directory_entries = (self._size_per_table + max_size - 1) // self._size_per_table

        self._aligned_table_size = (
            (self.block_size + 8 * self._num_table_entries - 1) // self.block_size * self.block_size
        )

        self.directories = sorted(
            (Directory(self, offset) for offset in self.header.directory_offsets),
            key=lambda d: d.version,
            reverse=True,
        )
        self.active_directory = self.directories[0]

        self.metadata_header = None
        self.metadata: dict[str, Any] = {}
        if self.header.metadata_chunk:
            # Open the file in reserved mode to read from the reserved area
            with self.open(reserved=True) as disk:
                metadata_offset = self.header.metadata_chunk * self.chunk_size
                disk.seek(metadata_offset)
                self.metadata_header = c_asif.asif_meta_header(disk)

                if self.metadata_header.header_signature != c_asif.ASIF_META_HEADER_SIGNATURE:
                    raise InvalidSignature(
                        f"Invalid a ASIF metadata header (expected {c_asif.ASIF_META_HEADER_SIGNATURE:#x}, "
                        f"got {self.metadata_header.header_signature:#x})"
                    )

                disk.seek(metadata_offset + self.metadata_header.header_size)
                self.metadata = plistlib.loads(disk.read(self.metadata_header.data_size).strip(b"\x00"))

    @property
    def internal_metadata(self) -> dict[str, Any]:
        """Get internal metadata from the ASIF image.

        Returns:
            A dictionary containing the internal metadata.
        """
        return self.metadata.get("internal metadata", {})

    @property
    def user_metadata(self) -> dict[str, Any]:
        """Get user metadata from the ASIF image.

        Returns:
            A dictionary containing the user metadata.
        """
        return self.metadata.get("user metadata", {})

    def open(self, reserved: bool = False) -> DataStream:
        """Open a stream to read the ASIF image data.

        Args:
            reserved: Whether to allow reading into the reserved area of the ASIF image.

        Returns:
            A stream-like object that can be used to read the image data.
        """
        return DataStream(self, reserved)


class Directory:
    """ASIF Directory.

    A directory has a version (``uint64``) followed by a list of table entries (``uint64[]``).
    The version number is used to determine the active directory, with the highest version being the active one.
    Each table entry is a chunk number and points to a table in the ASIF image.

    Args:
        asif: The ASIF image this directory belongs to.
        offset: Offset of the directory in the ASIF image.
    """

    def __init__(self, asif: ASIF, offset: int):
        self.asif = asif
        self.offset = offset

        self.asif.fh.seek(offset)
        self.version = c_asif.uint64(self.asif.fh)

        self.table = lru_cache(128)(self.table)

    def __repr__(self) -> str:
        return f"<Directory offset={self.offset:#x} version={self.version}>"

    @cached_property
    def entries(self) -> list[int]:
        """List of table entries in the directory."""
        # Seek over the version
        self.asif.fh.seek(self.offset + 8)
        return c_asif.uint64[self.asif._num_directory_entries](self.asif.fh)

    def table(self, index: int) -> Table:
        """Get a table from the directory.

        Args:
            index: Index of the table in the directory.
        """
        if index >= self.asif._num_directory_entries:
            raise IndexError("Table index out of range")
        return Table(self, index)


class Table:
    """ASIF Table.

    A table contains a list of data entries (``uint64[]``). Each data entry is a chunk number and points to a chunk of
    data in the ASIF image. Each table covers a fixed amount of data in the virtual disk.

    Data entries have 55 bits usable for the chunk number and 9 bits reserved for flags.

    .. rubric :: Encoding
    .. code-block:: c

        0b00000000 01111111 11111111 11111111 11111111 11111111 11111111 11111111  (chunk number)
        0b00111111 10000000 00000000 00000000 00000000 00000000 00000000 00000000  (reserved)
        0b01000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000  (entry dirty)
        0b10000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000  (content dirty)

    Args:
        directory: The directory this table belongs to.
        index: Index of the table in the directory.
    """

    def __init__(self, directory: Directory, index: int):
        self.asif = directory.asif
        self.directory = directory
        self.index = index

        self.offset = self.directory.entries[index] * self.asif.chunk_size
        self.virtual_offset = index * self.asif._size_per_table

    def __repr__(self) -> str:
        return f"<Table index={self.index} offset={self.offset:#x} virtual_offset={self.virtual_offset:#x}>"

    @cached_property
    def entries(self) -> list[int]:
        """List of data entries in the table."""
        self.asif.fh.seek(self.offset)
        return c_asif.uint64[self.asif._num_table_entries](self.asif.fh)


class DataStream(AlignedStream):
    """Stream to read data from an ASIF image.

    Args:
        asif: The ASIF image to read from.
        reserved: Whether to allow reading into the reserved area of the ASIF image.
    """

    def __init__(self, asif: ASIF, reserved: bool = False):
        super().__init__(asif.max_size if reserved else asif.size, align=asif.chunk_size)
        self.asif = asif
        self.reserved = reserved
        self.directory = asif.active_directory

    def _read(self, offset: int, length: int) -> bytes:
        result = []
        while length:
            table = self.directory.table(offset // self.asif._size_per_table)
            relative_block_index = (offset // self.asif.block_size) - (table.virtual_offset // self.asif.block_size)
            data_idx = (
                relative_block_index // self.asif._blocks_per_chunk
                + relative_block_index // self.asif._blocks_per_chunk * self.asif._num_reserved_table_entries
            ) // self.asif._num_reserved_table_entries

            # 0x8000000000000000 = content dirty bit
            # 0x4000000000000000 = entry dirty bit
            # 0x3F80000000000000 = reserved bits
            chunk = table.entries[data_idx] & 0x7FFFFFFFFFFFFF
            raw_offset = chunk * self.asif.chunk_size

            read_length = min(length, self.asif.chunk_size)
            if chunk == 0:
                result.append(b"\x00" * read_length)
            else:
                self.asif.fh.seek(raw_offset)
                result.append(self.asif.fh.read(read_length))

            offset += read_length
            length -= read_length

        return b"".join(result)



from __future__ import annotations

from dissect.hypervisor.descriptor import hyperv, ovf, pvs, vbox, vmx
from dissect.hypervisor.disk import asif, hdd, qcow2, vdi, vhd, vhdx, vmdk
from dissect.hypervisor.util import envelope, vmtar

__all__ = [
    "asif",
    "envelope",
    "hdd",
    "hyperv",
    "ovf",
    "pvs",
    "qcow2",
    "vbox",
    "vdi",
    "vhd",
    "vhdx",
    "vmdk",
    "vmtar",
    "vmx",
]



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_asif(__cs__.cstruct):
    ASIF_HEADER_SIGNATURE: Literal[1936221303] = ...
    ASIF_META_HEADER_SIGNATURE: Literal[1835365473] = ...
    class asif_header(__cs__.Structure):
        header_signature: _c_asif.uint32
        header_version: _c_asif.uint32
        header_size: _c_asif.uint32
        header_flags: _c_asif.uint32
        directory_offsets: __cs__.Array[_c_asif.uint64]
        guid: __cs__.CharArray
        sector_count: _c_asif.uint64
        max_sector_count: _c_asif.uint64
        chunk_size: _c_asif.uint32
        block_size: _c_asif.uint16
        total_segments: _c_asif.uint16
        metadata_chunk: _c_asif.uint64
        unk_50: __cs__.CharArray
        read_only_flags: _c_asif.uint32
        metadata_flags: _c_asif.uint32
        metadata_read_only_flags: _c_asif.uint32
        @overload
        def __init__(
            self,
            header_signature: _c_asif.uint32 | None = ...,
            header_version: _c_asif.uint32 | None = ...,
            header_size: _c_asif.uint32 | None = ...,
            header_flags: _c_asif.uint32 | None = ...,
            directory_offsets: __cs__.Array[_c_asif.uint64] | None = ...,
            guid: __cs__.CharArray | None = ...,
            sector_count: _c_asif.uint64 | None = ...,
            max_sector_count: _c_asif.uint64 | None = ...,
            chunk_size: _c_asif.uint32 | None = ...,
            block_size: _c_asif.uint16 | None = ...,
            total_segments: _c_asif.uint16 | None = ...,
            metadata_chunk: _c_asif.uint64 | None = ...,
            unk_50: __cs__.CharArray | None = ...,
            read_only_flags: _c_asif.uint32 | None = ...,
            metadata_flags: _c_asif.uint32 | None = ...,
            metadata_read_only_flags: _c_asif.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class asif_meta_header(__cs__.Structure):
        header_signature: _c_asif.uint32
        header_version: _c_asif.uint32
        header_size: _c_asif.uint32
        data_size: _c_asif.uint64
        unk_14: _c_asif.uint64
        @overload
        def __init__(
            self,
            header_signature: _c_asif.uint32 | None = ...,
            header_version: _c_asif.uint32 | None = ...,
            header_size: _c_asif.uint32 | None = ...,
            data_size: _c_asif.uint64 | None = ...,
            unk_14: _c_asif.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_asif` is an instance of `_c_asif`, but then we can't use it in type hints
c_asif: TypeAlias = _c_asif



from __future__ import annotations

import ctypes
import io
import logging
import os
import re
import textwrap
import zlib
from bisect import bisect_right
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, BinaryIO

from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_vmdk import (
    COWD_MAGIC,
    SECTOR_SIZE,
    SESPARSE_MAGIC,
    VMDK_MAGIC,
    c_vmdk,
)

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_VMDK", "CRITICAL"))


class VMDK(AlignedStream):
    def __init__(self, fh: BinaryIO | Path | str | list[BinaryIO | Path | str]):
        """Input can be a file handle to a Disk Descriptor file or a list of file handles to multiple VMDK files."""
        fhs = [fh] if not isinstance(fh, list) else fh

        self.disks = []
        self.parent = None
        self.descriptor = None
        self._disk_offsets = []
        self.sector_count = 0

        for fh in fhs:
            if hasattr(fh, "read"):
                name = getattr(fh, "name", None)
                path = Path(name) if name else None
            else:
                if not isinstance(fh, Path):
                    fh = Path(fh)
                path = fh
                fh = path.open("rb")

            magic = fh.read(4)
            fh.seek(-4, io.SEEK_CUR)

            if magic == b"# Di" and len(fhs) == 1:
                # Try reading the disk files from this descriptor
                # Otherwise we assume that the other file handles are the appropriate disks
                self.descriptor = DiskDescriptor.parse(fh.read().decode())

                if self.descriptor.attr["parentCID"] != "ffffffff":
                    self.parent = open_parent(path.parent, self.descriptor.attr["parentFileNameHint"])

                for extent in self.descriptor.extents:
                    if extent.type in ["SPARSE", "VMFSSPARSE", "SESPARSE"]:
                        sdisk_fh = path.with_name(extent.filename).open("rb")
                        self.disks.append(SparseDisk(sdisk_fh, parent=self.parent))
                    elif extent.type in ["VMFS", "FLAT"]:
                        rdisk_fh = path.with_name(extent.filename).open("rb")
                        self.disks.append(RawDisk(rdisk_fh, extent.sectors * SECTOR_SIZE))

            elif magic in (COWD_MAGIC, VMDK_MAGIC, SESPARSE_MAGIC):
                sparse_disk = SparseDisk(fh)

                if sparse_disk.descriptor and sparse_disk.descriptor.attr["parentCID"] != "ffffffff":
                    sparse_disk.parent = open_parent(path.parent, sparse_disk.descriptor.attr["parentFileNameHint"])
                self.disks.append(sparse_disk)

            else:
                self.disks.append(RawDisk(fh))

        size = 0
        for disk in self.disks:
            if size != 0:
                self._disk_offsets.append(self.sector_count)
            disk.offset = size
            disk.sector_offset = self.sector_count
            size += disk.size
            self.sector_count += disk.sector_count

        super().__init__(size)

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("VMDK::read_sectors(0x%x, 0x%x)", sector, count)

        sectors_read = []

        disk_idx = bisect_right(self._disk_offsets, sector)

        while count > 0:
            disk = self.disks[disk_idx]

            disk_remaining_sectors = disk.sector_count - (sector - disk.sector_offset)
            disk_sectors = min(disk_remaining_sectors, count)

            sectors_read.append(disk.read_sectors(sector, disk_sectors))

            sector += disk_sectors
            count -= disk_sectors
            disk_idx += 1

        return b"".join(sectors_read)

    def _read(self, offset: int, length: int) -> bytes:
        log.debug("VMDK::_read(0x%x, 0x%x)", offset, length)

        sector = offset // SECTOR_SIZE
        count = (length + SECTOR_SIZE - 1) // SECTOR_SIZE

        return self.read_sectors(sector, count)


class RawDisk:
    def __init__(self, fh: BinaryIO, size: int | None = None, offset: int = 0, sector_offset: int = 0):
        self.fh = fh
        self.offset = offset
        self.sector_offset = sector_offset

        if not size:
            fh.seek(0, io.SEEK_END)
            self.size = fh.tell()
            fh.seek(0)
        else:
            self.size = size

        self.sector_count = self.size // SECTOR_SIZE

        self.read = fh.read
        self.seek = fh.seek
        self.tell = fh.tell

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("RawDisk::read_sectors(0x%x)", sector)

        self.fh.seek((sector - self.sector_offset) * SECTOR_SIZE)
        return self.fh.read(count * SECTOR_SIZE)


class SparseDisk:
    def __init__(
        self, fh: BinaryIO, parent: VMDK | RawDisk | SparseDisk | None = None, offset: int = 0, sector_offset: int = 0
    ):
        self.fh = fh
        self.parent = parent
        self.offset = offset
        self.sector_offset = sector_offset

        fh.seek(0, io.SEEK_END)
        self.filesize = fh.tell()
        fh.seek(0, io.SEEK_SET)
        self.descriptor = None

        self.header = SparseExtentHeader(fh)
        if self.header.magic in (VMDK_MAGIC, COWD_MAGIC):
            self.is_sesparse = False

            if ctypes.c_int64(self.header.primary_grain_directory_offset).value == -1:
                # Footer starts -1024 from the end
                fh.seek(-1024, io.SEEK_END)
                self.header = SparseExtentHeader(fh)

            if self.header.magic == VMDK_MAGIC:
                grain_table_coverage = self.header.num_grain_table_entries * self.header.grain_size
                self._grain_directory_size = (self.header.capacity + grain_table_coverage - 1) // grain_table_coverage
                self._grain_table_size = self.header.num_grain_table_entries

                if self.header.descriptor_size > 0:
                    fh.seek(self.header.descriptor_offset * SECTOR_SIZE)
                    descriptor_buf = fh.read(self.header.descriptor_size * SECTOR_SIZE)
                    self.descriptor = DiskDescriptor.parse(descriptor_buf.split(b"\x00", 1)[0].decode())

            elif self.header.magic == COWD_MAGIC:
                self._grain_directory_size = self.header.num_grain_directory_entries
                self._grain_table_size = 4096

            grain_directory_offset = self.header.primary_grain_directory_offset
            self._grain_entry_type = c_vmdk.uint32

            self._grain_directory = c_vmdk.uint32[self._grain_directory_size](fh)

        elif self.header.magic == c_vmdk.SESPARSE_CONST_HEADER_MAGIC:
            self.is_sesparse = True

            self._grain_directory_size = self.header.grain_directory_size * SECTOR_SIZE // 8
            self._grain_table_size = self.header.grain_table_size * SECTOR_SIZE // 8

            grain_directory_offset = self.header.grain_directory_offset
            self._grain_entry_type = c_vmdk.uint64

        self.fh.seek(grain_directory_offset * SECTOR_SIZE)
        self._grain_directory = self._grain_entry_type[self._grain_directory_size](fh)

        self.size = self.header.capacity * SECTOR_SIZE
        self.sector_count = self.header.capacity

        self._lookup_grain_table = lru_cache(128)(self._lookup_grain_table)

    def _lookup_grain_table(self, directory: int) -> list[int]:
        gtbl_offset = self._grain_directory[directory]

        if self.is_sesparse:
            # qemu/block/vmdk.c:
            # Top most nibble is 0x1 if grain table is allocated.
            # strict check - top most 4 bytes must be 0x10000000 since max
            # supported size is 64TB for disk - so no more than 64TB / 16MB
            # grain directories which is smaller than uint32,
            # where 16MB is the only supported default grain table coverage.
            if not gtbl_offset or gtbl_offset & 0xFFFFFFFF00000000 != 0x1000000000000000:
                table = None
            else:
                gtbl_offset &= 0x00000000FFFFFFFF
                gtbl_offset = (
                    self.header.grain_tables_offset + gtbl_offset * (self._grain_table_size * 8) // SECTOR_SIZE
                )
                self.fh.seek(gtbl_offset * SECTOR_SIZE)
                table = self._grain_entry_type[self._grain_table_size](self.fh)
        else:
            if gtbl_offset:
                self.fh.seek(gtbl_offset * SECTOR_SIZE)
                table = self._grain_entry_type[self._grain_table_size](self.fh)
            else:
                table = None

        return table

    def _lookup_grain(self, grain: int) -> int:
        gdir_entry, gtbl_entry = divmod(grain, self._grain_table_size)
        table = self._lookup_grain_table(gdir_entry)

        if table:
            grain_entry = table[gtbl_entry]
            if self.is_sesparse:
                # SESparse uses a different method of specifying unallocated/sparse/allocated grains
                # However, we can re-use the normal sparse logic of returning 0 for unallocated, 1 for
                # sparse and >1 for allocated grains, since a grain of 0 or 1 isn't possible in SESparse.
                grain_type = grain_entry & c_vmdk.SESPARSE_GRAIN_TYPE_MASK
                if grain_type in (c_vmdk.SESPARSE_GRAIN_TYPE_UNALLOCATED, c_vmdk.SESPARSE_GRAIN_TYPE_FALLTHROUGH):
                    # Unallocated or scsi unmapped, fallthrough
                    return 0
                if grain_type == c_vmdk.SESPARSE_GRAIN_TYPE_ZERO:
                    # Sparse, zero grain
                    return 1
                if grain_type == c_vmdk.SESPARSE_GRAIN_TYPE_ALLOCATED:
                    # Allocated
                    cluster_sector_hi = (grain_entry & 0x0FFF000000000000) >> 48
                    cluster_sector_lo = (grain_entry & 0x0000FFFFFFFFFFFF) << 12
                    cluster_sector = cluster_sector_hi | cluster_sector_lo
                    return self.header.grains_offset + cluster_sector * self.header.grain_size

                raise ValueError("Unknown grain type")

            return grain_entry

        return 0

    def get_runs(self, sector: int, count: int) -> list[tuple[int, int, int, int | None]]:
        disk_sector = sector - self.sector_offset

        run_type = None
        run_offset = 0
        run_count = 0
        run_parent = None
        next_grain_sector = 0

        read_sector = disk_sector
        read_count = count

        runs = []

        if read_count == 0:
            return runs

        while read_count > 0:
            grain, grain_offset = divmod(read_sector, self.header.grain_size)
            grain_sector = self._lookup_grain(grain)
            read_sector_count = min(read_count, self.header.grain_size - grain_offset)

            if (run_type == 0 and grain_sector == 0) or (run_type == 1 and grain_sector == 1):
                run_count += read_sector_count
            elif run_type and run_type > 1 and grain_sector == next_grain_sector:
                next_grain_sector += self.header.grain_size
                run_count += read_sector_count
            else:
                if run_type is not None:
                    runs.append((run_type, run_offset, run_count, run_parent))
                    run_type = None
                    run_count = 0
                    run_parent = None
                if grain_sector == 0:
                    run_type = 0
                    run_count += read_sector_count
                    run_parent = self.sector_offset + read_sector
                elif grain_sector == 1:
                    run_type = 1
                    run_count += read_sector_count
                else:
                    run_type = grain_sector
                    run_offset = grain_offset
                    run_count += read_sector_count
                    next_grain_sector = grain_sector + self.header.grain_size

            read_count -= read_sector_count
            read_sector += read_sector_count

        assert run_type is not None
        runs.append((run_type, run_offset, run_count, run_parent))

        return runs

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("SparseDisk::read_sectors(0x%x, 0x%x)", sector, count)

        runs = self.get_runs(sector, count)
        sectors_read = []

        for run_type, run_offset, run_count, run_parent in runs:
            # Grain not present
            if run_type == 0:
                if self.parent:
                    sector_data = self.parent.read_sectors(run_parent, run_count)
                else:
                    sector_data = b"\x00" * (run_count * SECTOR_SIZE)
                sectors_read.append(sector_data)
                continue

            # Sparse grain
            if run_type == 1:
                sectors_read.append(b"\x00" * (run_count * SECTOR_SIZE))
                continue

            # Uncompressed grain
            if self.header.flags & c_vmdk.SPARSEFLAG_COMPRESSED == 0:
                self.fh.seek((run_type + run_offset) * SECTOR_SIZE)
                sector_data = self.fh.read(run_count * SECTOR_SIZE)
                sectors_read.append(sector_data)
                continue

            # Compressed grain
            while run_count > 0:
                # We consolidate grain runs in get_runs, but we can't read a contiguous stream of compressed grains
                # So loop over the consolidated grains
                offset = run_offset * SECTOR_SIZE
                grain_remaining = self.header.grain_size - run_offset
                read_count = min(run_count, grain_remaining)

                buf = self._read_compressed_grain(run_type)
                sectors_read.append(buf[offset : offset + read_count * SECTOR_SIZE])

                # If we loop, we're going to the next run, which means we'll start at offset 0
                run_offset = 0
                run_type += self.header.grain_size
                run_count -= read_count

        return b"".join(sectors_read)

    def _read_compressed_grain(self, sector: int) -> bytes:
        self.fh.seek(sector * SECTOR_SIZE)
        buf = self.fh.read(SECTOR_SIZE)

        if self.header.flags & c_vmdk.SPARSEFLAG_EMBEDDED_LBA:
            header_len = 12
            lba_header = c_vmdk.SparseGrainLBAHeaderOnDisk(buf)
            compressed_len = lba_header.cmp_size
        else:
            header_len = 4
            compressed_len = c_vmdk.uint32(buf)

        if compressed_len + header_len > SECTOR_SIZE:
            # Officially this is padded to SECTOR_SIZE, but we don't really care
            remaining_len = header_len + compressed_len - SECTOR_SIZE
            self.fh.seek((sector + 1) * SECTOR_SIZE)
            buf += self.fh.read(remaining_len)

        return zlib.decompress(buf[header_len : header_len + compressed_len])


class SparseExtentHeader:
    def __init__(self, fh: BinaryIO):
        magic = fh.read(4)
        fh.seek(-4, io.SEEK_CUR)

        if magic == VMDK_MAGIC:
            self.hdr = c_vmdk.VMDKSparseExtentHeader(fh)
        elif magic == SESPARSE_MAGIC:
            self.hdr = c_vmdk.VMDKSESparseConstHeader(fh)
        elif magic == COWD_MAGIC:
            self.hdr = c_vmdk.COWDSparseExtentHeader(fh)
        else:
            raise NotImplementedError("Unsupported sparse extent")

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.hdr, attr)


RE_EXTENT_DESCRIPTOR = re.compile(
    r"""
    ^
    (?P<access_mode>RW|RDONLY|NOACCESS)\s
    (?P<sectors>\d+)\s
    (?P<type>SESPARSE|SPARSE|ZERO|FLAT|VMFS|VMFSSPARSE|VMFSRDM|VMFSRAW)
    (\s(?P<filename>\".+\"))?
    (\s(?P<start_sector>\d+))?
    (\s(?P<partition_uuid>\S+))?
    (\s(?P<device_identifier>\S+))?
    $
    """,
    re.VERBOSE,
)


@dataclass
class ExtentDescriptor:
    raw: str
    access_mode: str
    sectors: int
    type: str
    filename: str | None = None
    start_sector: int | None = None
    partition_uuid: str | None = None
    device_identifier: str | None = None

    def __post_init__(self) -> None:
        self.sectors = int(self.sectors)

        if self.filename:
            self.filename = self.filename.strip('"')

        if self.start_sector:
            self.start_sector = int(self.start_sector)

    def __repr__(self) -> str:
        return f"<ExtentDescriptor {self.raw}>"

    def __str__(self) -> str:
        return self.raw


class DiskDescriptor:
    def __init__(
        self, attr: dict, extents: list[ExtentDescriptor], disk_db: dict, sectors: int, raw_config: str | None = None
    ):
        self.attr = attr
        self.extents = extents
        self.ddb = disk_db
        self.sectors = sectors
        self.raw = raw_config

    @classmethod
    def parse(cls, vmdk_config: str) -> DiskDescriptor:
        """Return :class:`DiskDescriptor` based on the provided ``vmdk_config``.

        Resources:
            - https://github.com/libyal/libvmdk/blob/main/documentation/VMWare%20Virtual%20Disk%20Format%20(VMDK).asciidoc
        """
        descriptor_settings = {}
        extents: list[ExtentDescriptor] = []
        disk_db = {}
        sectors = 0

        for line in vmdk_config.split("\n"):
            line = line.strip()

            if not line or line.startswith("#"):
                continue

            if line.startswith(("RW ", "RDONLY ", "NOACCESS ")):
                match = RE_EXTENT_DESCRIPTOR.search(line)

                if not match:
                    log.warning("Unexpected ExtentDescriptor format in vmdk config: %s, ignoring", line)
                    continue

                extent = ExtentDescriptor(raw=line, **match.groupdict())
                sectors += extent.sectors
                extents.append(extent)
                continue

            setting, _, value = line.partition("=")
            setting = setting.strip()
            value = value.strip(' "')

            if setting.startswith("ddb."):
                disk_db[setting] = value
            else:
                descriptor_settings[setting] = value

        return cls(descriptor_settings, extents, disk_db, sectors, vmdk_config)

    def __str__(self) -> str:
        str_template = textwrap.dedent(
            """\
        # Disk DescriptorFile
        version=1
        {}

        # Extent Description
        {}

        # The Disk Data Base
        #DDB

        {}"""
        )

        descriptor_settings = []
        for setting, value in self.attr.items():
            if setting != "version":
                descriptor_settings.append(f"{setting}={value}")
        descriptor_settings = "\n".join(descriptor_settings)

        extents = "\n".join(map(str, self.extents))

        disk_db = []
        for setting, value in self.ddb.items():
            disk_db.append(f'{setting} = "{value}"')
        disk_db = "\n".join(disk_db)

        return str_template.format(descriptor_settings, extents, disk_db)


def open_parent(path: Path, filename_hint: str) -> VMDK:
    try:
        filename_hint = filename_hint.replace("\\", "/")
        hint_path, _, filename = filename_hint.rpartition("/")
        filepath = path.joinpath(filename)
        if not filepath.exists():
            filepath = path.joinpath(filename_hint)
            if not filepath.exists():
                _, _, hint_path_name = hint_path.rpartition("/")
                filepath = path.parent.joinpath(hint_path_name).joinpath(filename)
        vmdk = VMDK(filepath)
    except Exception as err:
        raise IOError(f"Failed to open parent disk with hint {filename_hint} from path {path}: {err}")

    return vmdk



from __future__ import annotations

from typing import TYPE_CHECKING, Final, TextIO

from defusedxml import ElementTree

if TYPE_CHECKING:
    from collections.abc import Iterator
    from xml.etree.ElementTree import Element


class OVF:
    NS: Final[dict[str, str]] = {
        "ovf": "http://schemas.dmtf.org/ovf/envelope/1",
        "rasd": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2/CIM_ResourceAllocationSettingData",
    }

    FILE_XPATH = "ovf:References/ovf:File"
    DISK_XPATH = "ovf:DiskSection/ovf:Disk"
    DISK_DRIVE_XPATH = 'ovf:VirtualSystem/ovf:VirtualHardwareSection/ovf:Item/[rasd:ResourceType="17"]'

    def __init__(self, fh: TextIO):
        self.fh = fh
        self.xml: Element = ElementTree.fromstring(fh.read())

        self.references = {}
        for file_ in self.xml.findall(self.FILE_XPATH, self.NS):
            file_id = file_.get("{{{ovf}}}id".format(**self.NS))
            href = file_.get("{{{ovf}}}href".format(**self.NS))
            self.references[file_id] = href

        self._disks = {}
        for disk in self.xml.findall(self.DISK_XPATH, self.NS):
            disk_id = disk.get("{{{ovf}}}diskId".format(**self.NS))
            file_ref = disk.get("{{{ovf}}}fileRef".format(**self.NS))
            self._disks[disk_id] = self.references[file_ref]

    def disks(self) -> Iterator[str]:
        for disk in self.xml.findall(self.DISK_DRIVE_XPATH, self.NS):
            resource = disk.find("{{{rasd}}}HostResource".format(**self.NS))
            xpath = resource.text
            xpath = xpath.removeprefix("ovf:")

            if xpath.startswith("/disk/"):
                disk_ref = xpath.split("/")[-1]
                yield self._disks[disk_ref]
            elif xpath.startswith("/file/"):
                file_ref = xpath.split("/")[-1]
                yield self.references[file_ref]
            else:
                raise NotImplementedError(resource)



from __future__ import annotations

from dissect.cstruct import cstruct

qcow2_def = """
#define MIN_CLUSTER_BITS 9
#define MAX_CLUSTER_BITS 21

#define QCOW2_COMPRESSED_SECTOR_SIZE    512
#define QCOW2_COMPRESSION_TYPE_ZLIB     0
#define QCOW2_COMPRESSION_TYPE_ZSTD     1

#define L1E_SIZE            8   // (sizeof(uint64_t))

#define L2E_SIZE_NORMAL     8   // (sizeof(uint64_t))
#define L2E_SIZE_EXTENDED   16  // (sizeof(uint64_t) * 2)

#define L1E_OFFSET_MASK 0x00fffffffffffe00
#define L2E_OFFSET_MASK 0x00fffffffffffe00
#define L2E_COMPRESSED_OFFSET_SIZE_MASK 0x3fffffffffffffff

/* indicate that the refcount of the referenced cluster is exactly one. */
#define QCOW_OFLAG_COPIED       (1 << 63)
/* indicate that the cluster is compressed (they never have the copied flag) */
#define QCOW_OFLAG_COMPRESSED   (1 << 62)
/* The cluster reads as all zeros */
#define QCOW_OFLAG_ZERO         (1 << 0)

#define QCOW_EXTL2_SUBCLUSTERS_PER_CLUSTER 32

#define QCOW2_INCOMPAT_DIRTY_BITNR          0
#define QCOW2_INCOMPAT_CORRUPT_BITNR        1
#define QCOW2_INCOMPAT_DATA_FILE_BITNR      2
#define QCOW2_INCOMPAT_COMPRESSION_BITNR    3
#define QCOW2_INCOMPAT_EXTL2_BITNR          4
#define QCOW2_INCOMPAT_DIRTY                1 << QCOW2_INCOMPAT_DIRTY_BITNR
#define QCOW2_INCOMPAT_CORRUPT              1 << QCOW2_INCOMPAT_CORRUPT_BITNR
#define QCOW2_INCOMPAT_DATA_FILE            1 << QCOW2_INCOMPAT_DATA_FILE_BITNR
#define QCOW2_INCOMPAT_COMPRESSION          1 << QCOW2_INCOMPAT_COMPRESSION_BITNR
#define QCOW2_INCOMPAT_EXTL2                1 << QCOW2_INCOMPAT_EXTL2_BITNR

typedef struct {
    uint32_t magic;
    uint32_t version;
    uint64_t backing_file_offset;
    uint32_t backing_file_size;
    uint32_t cluster_bits;
    uint64_t size; /* in bytes */
    uint32_t crypt_method;
    uint32_t l1_size; /* XXX: save number of clusters instead ? */
    uint64_t l1_table_offset;
    uint64_t refcount_table_offset;
    uint32_t refcount_table_clusters;
    uint32_t nb_snapshots;
    uint64_t snapshots_offset;

    /* The following fields are only valid for version >= 3 */
    uint64_t incompatible_features;
    uint64_t compatible_features;
    uint64_t autoclear_features;

    uint32_t refcount_order;
    uint32_t header_length;

    /* Additional fields */
    uint8_t compression_type;

    /* header must be a multiple of 8 */
    uint8_t padding[7];
} QCowHeader;

#define QCOW2_EXT_MAGIC_END             0
#define QCOW2_EXT_MAGIC_BACKING_FORMAT  0xe2792aca
#define QCOW2_EXT_MAGIC_FEATURE_TABLE   0x6803f857
#define QCOW2_EXT_MAGIC_CRYPTO_HEADER   0x0537be77
#define QCOW2_EXT_MAGIC_BITMAPS         0x23852875
#define QCOW2_EXT_MAGIC_DATA_FILE       0x44415441

typedef struct {
    uint32_t magic;
    uint32_t len;
} QCowExtension;

typedef struct {
    uint64_t offset;
    uint64_t length;
} Qcow2CryptoHeaderExtension;

typedef struct {
    uint32_t nb_bitmaps;
    uint32_t reserved32;
    uint64_t bitmap_directory_size;
    uint64_t bitmap_directory_offset;
} Qcow2BitmapHeaderExt;

typedef struct {
    /* header is 8 byte aligned */
    uint64_t l1_table_offset;

    uint32_t l1_size;
    uint16_t id_str_size;
    uint16_t name_size;

    uint32_t date_sec;
    uint32_t date_nsec;

    uint64_t vm_clock_nsec;

    uint32_t vm_state_size;
    uint32_t extra_data_size; /* for extension */
    /* extra data follows */
    /* id_str follows */
    /* name follows  */
} QCowSnapshotHeader;

typedef struct {
    uint64_t vm_state_size_large;
    uint64_t disk_size;
    uint64_t icount;
} QCowSnapshotExtraData;

enum QCow2ClusterType {
    QCOW2_CLUSTER_UNALLOCATED,
    QCOW2_CLUSTER_ZERO_PLAIN,
    QCOW2_CLUSTER_ZERO_ALLOC,
    QCOW2_CLUSTER_NORMAL,
    QCOW2_CLUSTER_COMPRESSED,
};

enum QCow2SubclusterType {
    QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN,
    QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC,
    QCOW2_SUBCLUSTER_ZERO_PLAIN,
    QCOW2_SUBCLUSTER_ZERO_ALLOC,
    QCOW2_SUBCLUSTER_NORMAL,
    QCOW2_SUBCLUSTER_COMPRESSED,
    QCOW2_SUBCLUSTER_INVALID,
};
"""

c_qcow2 = cstruct(endian=">").load(qcow2_def)

QCOW2_MAGIC = 0x514649FB
QCOW2_MAGIC_BYTES = c_qcow2.uint32.dumps(QCOW2_MAGIC)

QCOW2_INCOMPAT_MASK = (
    c_qcow2.QCOW2_INCOMPAT_DIRTY
    | c_qcow2.QCOW2_INCOMPAT_CORRUPT
    | c_qcow2.QCOW2_INCOMPAT_DATA_FILE
    | c_qcow2.QCOW2_INCOMPAT_COMPRESSION
    | c_qcow2.QCOW2_INCOMPAT_EXTL2
)

QCow2ClusterType = c_qcow2.QCow2ClusterType
QCow2SubclusterType = c_qcow2.QCow2SubclusterType

NORMAL_SUBCLUSTER_TYPES = (
    QCow2SubclusterType.QCOW2_SUBCLUSTER_NORMAL,
    QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_ALLOC,
    QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC,
)

ZERO_SUBCLUSTER_TYPES = (
    QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_PLAIN,
    QCow2SubclusterType.QCOW2_SUBCLUSTER_ZERO_ALLOC,
)

UNALLOCATED_SUBCLUSTER_TYPES = (
    QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_PLAIN,
    QCow2SubclusterType.QCOW2_SUBCLUSTER_UNALLOCATED_ALLOC,
)


def ctz(value: int) -> int:
    """Count the number of zero bits in an integer of a given size."""
    return (value & -value).bit_length() - 1 if value else 0



# References:
# - VmDataStore.dll
from __future__ import annotations

import struct
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import RangeStream

from dissect.hypervisor.descriptor.c_hyperv import (
    KeyDataFlag,
    KeyDataType,
    ObjectEntryType,
    c_hyperv,
)
from dissect.hypervisor.exceptions import InvalidSignature

if TYPE_CHECKING:
    from collections.abc import ItemsView, KeysView, ValuesView


class HyperVFile:
    """HyperVFile implementation.

    I think, technically, the underlying container is called ``HyperVStorage``, and the ``HyperVFile`` adds
    some features on top of that. We just call it ``HyperVFile`` for convenience.

    A ``HyperVFile`` has 2 headers, one at ``0x0000`` and one at ``0x1000``. The active header is determined
    by a sequence number.

    A replay log is located at an offset specified in the header. This replay log functions as a journal
    for changes made to the file. A file is dirty if it contains outstanding entries in the replay log,
    and replaying the specified log entries is necessary. This is not currently implemented.

    An object table seems to be located at ``0x2000``, but it's unclear if this is always the case. The offset
    might be related to the header size, log size or data alignment. This will need some more research.
    This table contains entries that describe the various "objects" contained within this file. The available
    types are listed in the ``ObjectEntryType`` enum.

    ``HyperVFile``'s have a version number, and it looks like there are some slight differences between
    different versions. The key tables are at least stored in a different manner, because there's code
    to handle loading them differently, and also to update them to the new format if an old version is
    encountered. However, I haven't seen any files with older versions yet, so we guard this implementation
    to only version ``0x0400`` at this time.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self.fh.seek(c_hyperv.FIRST_HEADER_OFFSET)
        header1 = c_hyperv.HyperVStorageHeader(self.fh)
        self.fh.seek(c_hyperv.SECOND_HEADER_OFFSET)
        header2 = c_hyperv.HyperVStorageHeader(self.fh)

        self.header = header1 if header1.sequence_number > header2.sequence_number else header2
        self.headers = [header1, header2]

        if self.header.signature != c_hyperv.SIGNATURE_STORAGE_HEADER:
            raise InvalidSignature(f"Invalid header signature: 0x{self.header.signature:x}")

        if self.header.version != 0x400:
            # Need test files to know how the keys are stored.
            raise NotImplementedError(f"Unsupported version: 0x{self.header.version:x}")

        self.replay_logs = [HyperVStorageReplayLog(self, self.header.replay_log_offset)]
        self.object_tables = [HyperVStorageObjectTable(self, c_hyperv.OBJECT_TABLE_OFFSET)]
        self.key_tables: dict[int, list[HyperVStorageKeyTable]] = {}
        self.file_objects: dict[int, HyperVStorageFileObject] = {}

        for object_table in self.object_tables:
            for entry in object_table.entries:
                if entry.allocated == 0:
                    continue

                if entry.type == ObjectEntryType.ObjectTable:
                    # Haven't seen a file yet with additional object tables, but I assume this is how it'd work
                    new_object_table = HyperVStorageObjectTable(self, entry.offset)
                    self.object_tables.append(new_object_table)

                if entry.type == ObjectEntryType.KeyTable:
                    key_table = HyperVStorageKeyTable(self, entry.offset, entry.size)
                    if key_table.index not in self.key_tables:
                        self.key_tables[key_table.index] = []
                    # Haven't seen a file yet with multiple tables of the same index, but different
                    # sequence number, but I assume this is how it'd work.
                    self.key_tables[key_table.index].append(key_table)
                    self.key_tables[key_table.index].sort(key=lambda table: table.sequence_number, reverse=True)

                if entry.type == ObjectEntryType.File:
                    self.file_objects[entry.offset] = HyperVStorageFileObject(self, entry.offset, entry.size)

                if entry.type == ObjectEntryType.ReplayLog:
                    # Haven't seen a file yet with additional replay logs, but I assume this is how it'd work
                    replay_log = HyperVStorageReplayLog(self, entry.offset)
                    self.replay_logs.append(replay_log)

        self.root = {}
        # Link all the children to their parents
        for key_tables in self.key_tables.values():
            active_table = key_tables[0]
            for entry in active_table.entries:
                if entry.type == KeyDataType.Free:
                    continue

                parent = entry.parent
                if parent:
                    parent.children[entry.key] = entry
                else:
                    self.root[entry.key] = entry

    def __getitem__(self, key: str) -> HyperVStorageKeyTableEntry:
        return self.root[key]

    def keys(self) -> KeysView[str]:
        return self.root.keys()

    def items(self) -> ItemsView[str, HyperVStorageKeyTableEntry]:
        return self.root.items()

    def values(self) -> ValuesView[HyperVStorageKeyTableEntry]:
        return self.root.values()

    def as_dict(self) -> dict:
        obj = {}

        for key, entry in self.root.items():
            obj[key] = entry.as_dict()

        return obj

    @property
    def version(self) -> int:
        return self.header.version

    def _align(self, offset: int) -> int:
        remainder = offset % self.header.alignment
        if remainder == 0:
            return offset
        return offset + self.header.alignment - remainder


class HyperVStorageReplayLog:
    """The replay log tracks changes in the file.

    Old changes are actually still resident in the log, but not counted in the ``num_entries`` field.
    """

    def __init__(self, hyperv_file: HyperVFile, offset: int):
        self.file = hyperv_file
        self.offset = offset

        self.file.fh.seek(offset)
        self.header = c_hyperv.HyperVStorageReplayLog(self.file.fh)

        if self.header.signature != c_hyperv.SIGNATURE_REPLAY_LOG_HEADER:
            raise InvalidSignature(f"Invalid replay log signature: 0x{self.header.signature:x}")

        self.entries = c_hyperv.HyperVStorageReplayLogEntry[self.header.num_entries](self.file.fh)


class HyperVStorageObjectTable:
    """The object table tracks all "objects".

    Objects are specific blocks in the file that have been designated a specific type.
    For example, the object table can list one or more key tables.
    """

    def __init__(self, hyperv_file: HyperVFile, offset: int):
        self.file = hyperv_file
        self.offset = offset

        self.file.fh.seek(offset)
        self.header = c_hyperv.HyperVStorageObjectTable(self.file.fh)

        if self.header.signature != c_hyperv.SIGNATURE_OBJECT_TABLE_HEADER:
            raise InvalidSignature(f"Invalid object table signature: 0x{self.header.signature:x}")

        self.entries = c_hyperv.HyperVStorageObjectTableEntry[self.header.num_entries](self.file.fh)


class HyperVStorageKeyTable:
    """The key table stores key value pairs and their relation to parent keys.

    A table has a specific table index and one or more key entries.
    """

    def __init__(self, hyperv_file: HyperVFile, offset: int, size: int):
        self.file = hyperv_file
        self.offset = offset
        self.size = size

        fh = self.file.fh
        fh.seek(offset)
        self.raw = memoryview(fh.read(size))

        self.header = c_hyperv.HyperVStorageKeyTable(self.raw)

        if self.header.signature != c_hyperv.SIGNATURE_KEY_TABLE_HEADER:
            raise InvalidSignature(f"Invalid key table signature: 0x{self.header.signature:x}")

        self.entries: list[HyperVStorageKeyTableEntry] = []
        self._lookup: dict[int, HyperVStorageKeyTableEntry] = {}

        entry_offset = len(c_hyperv.HyperVStorageKeyTable)
        while entry_offset < size:
            entry = HyperVStorageKeyTableEntry(self, entry_offset)
            if entry.size == 0:
                break

            self.entries.append(entry)
            self._lookup[entry_offset] = entry

            entry_offset = entry_offset + entry.size

    @property
    def index(self) -> int:
        """Return the table index."""
        return self.header.index

    @property
    def sequence_number(self) -> int:
        """Return the table sequence number."""
        return self.header.sequence_number


class HyperVStorageKeyTableEntry:
    """Entry in a key table.

    The first 16 bytes are a combined flag and type field. The high 8 bits are flags, the low 8 bits are type.

    Only one flag is currently known, which we've called ``FileObjectPointer``. It's the lowest bit of the flag bits.
    If this flag is set, it means the value of this entry is located in a file object. File objects pointers are
    12 bytes and consist of a ``uint32`` size and a ``uint64`` offset value. The offset is the absolute offset into the
    file. This method is similar to how parent keys are referenced.

    Values are stored in a file object if their size >= ``0x800``. As only strings and "arrays" are of variable size,
    these are the only data types that can be stored in file objects.

    Data type summary:

    - ``KeyDataType.Free``:
        - Allocated but free.

    - ``KeyDataType.Unknown``:
        - Unknown entry type. Anything greater than 9 and exactly 2 is unknown.

    - ``KeyDataType.Int``:
        - Signed 64 bit integer.

    - ``KeyDataType.UInt``:
        - Unsigned 64 bit integer.

    - ``KeyDataType.Double``:
        - 64 bit double.

    - ``KeyDataType.String``:
        - UTF-16-LE encoded string

    - ``KeyDataType.Array``:
        - Bytes?

    - ``KeyDataType.Bool``:
        - Boolean encoded as 32 bit integer.

    - ``KeyDataType.Node``:
        - Tree nodes. Value size is coded as 8, but actual size is 12.
        - First 8 bytes are unknown, but latter 4 bytes is the insertion sequence number.
    """

    def __init__(self, table: HyperVStorageKeyTable, offset: int):
        self.table = table
        self.offset = offset
        self.header = c_hyperv.HyperVStorageKeyTableEntryHeader(table.raw[offset:])
        self.children: dict[str, HyperVStorageKeyTableEntry] = {}

    def __getitem__(self, key: str) -> HyperVStorageKeyTableEntry:
        return self.children[key]

    def __repr__(self) -> str:
        return f"<HyperVStorageKeyTableEntry type={self.type} size={self.size}>"

    @property
    def parent(self) -> HyperVStorageKeyTableEntry | None:
        """Return the entry parent, if there is any.

        Requires that all key tables are loaded.
        """
        if not self.header.parent_table_idx:
            return None

        return self.table.file.key_tables[self.header.parent_table_idx][0]._lookup[self.header.parent_offset]

    @property
    def flags(self) -> int:
        """Return the entry flags."""
        return (self.header.type & 0xFF00) >> 8

    @property
    def type(self) -> KeyDataType:
        """Return the entry type."""
        return KeyDataType(self.header.type & 0xFF)

    @property
    def size(self) -> int:
        """Return the entry size."""
        return self.header.size

    @property
    def is_file_object_pointer(self) -> bool:
        """Return whether the value is a file object pointer."""
        return bool(KeyDataFlag(self.flags) & KeyDataFlag.FileObjectPointer)

    @property
    def file_object_pointer(self) -> tuple[int, int]:
        """Return the file object pointer information."""
        if not self.is_file_object_pointer:
            raise TypeError(f"KeyTableEntry is not a file object pointer: {self}")

        data = self.raw[self.header.data_offset :]
        size, offset = struct.unpack("<IQ", data[:12])
        # We swap the values since that is a bit more logical
        return offset, size

    @property
    def raw(self) -> memoryview:
        """Returns the raw data for this entry."""
        data_offset = self.offset + len(c_hyperv.HyperVStorageKeyTableEntryHeader)
        return self.table.raw[data_offset : self.offset + self.size]

    @property
    def data(self) -> memoryview:
        """Returns the data portion for this entry."""
        if self.is_file_object_pointer:
            _, size = self.file_object_pointer
            file_object = self.get_file_object()
            # This memoryview has no purpose, only do it so the return value type is consistent
            return memoryview(file_object.read(size))

        return self.raw[self.header.data_offset :]

    @property
    def key(self) -> str:
        """Returns the key name for this entry."""
        # Subtract 1 for the terminating null byte
        return self.raw.tobytes()[: self.header.data_offset - 1].decode("utf-8")

    @property
    def value(self) -> int | bytes | str:
        """Return a Python native value for this entry."""
        data = self.data

        if self.type == KeyDataType.Int:
            return struct.unpack("<q", data[:8])[0]

        if self.type == KeyDataType.UInt:
            return struct.unpack("<Q", data[:8])[0]

        if self.type == KeyDataType.Double:
            return struct.unpack("<d", data[:8])[0]

        if self.type in (KeyDataType.String, KeyDataType.Array):
            if not self.is_file_object_pointer:
                data_len = struct.unpack("<I", data[:4])[0]
                data = data[4 : 4 + data_len]

            data = data.tobytes()
            if self.type == KeyDataType.String:
                return data.decode("utf-16-le")
            return data

        if self.type == KeyDataType.Bool:
            return struct.unpack("<I", data[:4])[0] != 0

        raise TypeError(f"Unknown data type: {self.type}")

    @property
    def data_size(self) -> int:
        """Return the total amount of data bytes, including the key name.

        Reference:
            - ``HyperVStorageKeyTableEntry::GetDataSizeInBytes``
        """
        if self.type == KeyDataType.Node:
            return self.header.data_offset + 12
        return self.header.data_offset + self.value_size

    @property
    def value_size(self) -> int:
        """Return the amount of bytes a value occupies.

        Reference:
            - ``HyperVStorageKeyTableEntry::GetValueSizeInBytes``
        """
        if self.type == KeyDataType.Free:
            return self.header.size - len(c_hyperv.HyperVStorageKeyTableEntryHeader)

        if self.is_file_object_pointer:
            return 12

        if self.type == KeyDataType.Node:
            # The code returns 8 here, but it's actually 12?
            # First 8 bytes is an unknown value, but the latter 4 bytes are an insertion sequence number
            return 8

        if self.type in (KeyDataType.String, KeyDataType.Array):
            return struct.unpack("<I", self.raw[self.header.data_offset : self.header.data_offset + 4])[0]

        if self.type in (KeyDataType.Int, KeyDataType.UInt):
            return 8

        if self.type == KeyDataType.Double:
            return 8

        if self.type == KeyDataType.Bool:
            return 4

        raise TypeError(f"Unknown data type: {self.type}")

    def keys(self) -> KeysView[int]:
        return self.children.keys()

    def items(self) -> ItemsView[int, HyperVStorageKeyTableEntry]:
        return self.children.items()

    def values(self) -> ValuesView[HyperVStorageKeyTableEntry]:
        return self.children.values()

    def as_dict(self) -> dict:
        if self.type != KeyDataType.Node:
            raise TypeError(f"KeyTableEntry can't be dumped as dictionary: {self.type}")

        obj = {}
        for key, child in self.children.items():
            value = child.as_dict() if child.type == KeyDataType.Node else child.value

            obj[key] = value

        return obj

    def get_file_object(self) -> HyperVStorageFileObject:
        if not self.is_file_object_pointer:
            raise TypeError(f"Entry is not a file object pointer: {self}")

        offset, size = self.file_object_pointer
        file_object = self.table.file.file_objects.get(offset)
        if not file_object:
            raise ValueError(f"Unknown file object: 0x{offset:x} (0x{size:x})")

        return file_object


class HyperVStorageFileObject:
    """File object from the object table.

    File objects are referenced by their absolute offset in the file. The object table also stores
    a size, but this size will always be aligned with the data alignment of the Hyper-V file.
    The actual size of the data stored is located in the data portion of the ``HyperVStorageKeyTableEntry``
    that references that file object.
    """

    def __init__(self, hyperv_file: HyperVFile, offset: int, size: int):
        self.file = hyperv_file
        self.offset = offset
        self.size = size

    def read(self, n: int = -1) -> bytes:
        if n is not None and n < -1:
            raise ValueError("invalid number of bytes to read")

        read_length = self.size if n == -1 else min(n, self.size)

        self.file.fh.seek(self.offset)
        return self.file.fh.read(read_length)

    def open(self, size: int | None = None) -> BinaryIO:
        return RangeStream(self.file.fh, self.offset, size or self.size)



# References:
# - [MS-VHDX] https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-vhdx/83e061f8-f6e2-4de1-91bd-5d518a43d477

from __future__ import annotations

import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, Final
from uuid import UUID

from dissect.util.stream import AlignedStream

from dissect.hypervisor.disk.c_vhdx import (
    ALIGNMENT,
    BAT_REGION_GUID,
    FILE_PARAMETERS_GUID,
    LOGICAL_SECTOR_SIZE_GUID,
    MB,
    METADATA_REGION_GUID,
    PARENT_LOCATOR_GUID,
    PHYSICAL_SECTOR_SIZE_GUID,
    VHDX_PARENT_LOCATOR_GUID,
    VIRTUAL_DISK_ID_GUID,
    VIRTUAL_DISK_SIZE_GUID,
    c_vhdx,
)
from dissect.hypervisor.exceptions import InvalidSignature, InvalidVirtualDisk

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_VHDX", "CRITICAL"))


class VHDX(AlignedStream):
    """Hyper-V VHDX implementation.

    Supports fixed, dynamic and differencing VHDX files.

    Currently, differencing VHDX files must be passed as a pathlib.Path object, with
    the parent VHDX in the same directory, or the registered absolute directory.
    """

    def __init__(self, fh: BinaryIO | Path | str):
        if hasattr(fh, "read"):
            name = getattr(fh, "name", None)
            path = Path(name) if name else None
        else:
            if not isinstance(fh, Path):
                fh = Path(fh)
            path = fh
            fh = path.open("rb")

        self.fh = fh
        self.path = path

        self.file_identifier = c_vhdx.file_identifier(fh)
        if self.file_identifier.signature != b"vhdxfile":
            raise InvalidSignature(f"Invalid file identifier signature: {self.file_identifier.signature}")

        fh.seek(1 * ALIGNMENT)
        header1 = c_vhdx.header(fh)
        fh.seek(2 * ALIGNMENT)
        header2 = c_vhdx.header(fh)

        self.header = header1 if header1.sequence_number > header2.sequence_number else header2
        self.headers = [header1, header2]

        if self.header.signature != b"head":
            raise InvalidSignature(f"Invalid header signature: {self.header.signature}")

        region_table1 = RegionTable(fh, 3 * ALIGNMENT)
        region_table2 = RegionTable(fh, 4 * ALIGNMENT)

        self.region_table = region_table1
        self.region_tables = [region_table1, region_table2]

        metadata_entry = self.region_table.get(METADATA_REGION_GUID)
        self.metadata = MetadataTable(fh, metadata_entry.file_offset, metadata_entry.length)

        self.size = self.metadata.get(VIRTUAL_DISK_SIZE_GUID)
        file_parameters = self.metadata.get(FILE_PARAMETERS_GUID)
        self.block_size = file_parameters.block_size
        self.has_parent = file_parameters.has_parent
        self.sector_size = self.metadata.get(LOGICAL_SECTOR_SIZE_GUID)
        self.id = UUID(bytes_le=self.metadata.get(VIRTUAL_DISK_ID_GUID).virtual_disk_id)
        self._sectors_per_block = self.block_size // self.sector_size
        self._chunk_ratio = ((2**23) * self.sector_size) // self.block_size

        self.parent = None
        self.parent_locator: ParentLocator = None
        if self.has_parent:
            self.parent_locator = self.metadata.get(PARENT_LOCATOR_GUID)
            if self.parent_locator.type != VHDX_PARENT_LOCATOR_GUID:
                raise ValueError(f"Unknown parent locator type: {self.parent_locator.type}")
            self.parent = open_parent(self.path.parent, self.parent_locator.entries)

        bat_entry = self.region_table.get(BAT_REGION_GUID)
        self.bat = BlockAllocationTable(self, bat_entry.file_offset)

        super().__init__(self.size)

    def read_sectors(self, sector: int, count: int) -> bytes:
        log.debug("VHDX::read_sectors(0x%x, 0x%x)", sector, count)
        sectors_read = []

        while count > 0:
            read_count = min(count, self._sectors_per_block)
            read_size = read_count * self.sector_size
            block, sector_in_block = divmod(sector, self._sectors_per_block)
            bat_entry = self.bat.pb(block)

            if bat_entry.state == c_vhdx.PAYLOAD_BLOCK_NOT_PRESENT:
                # The block is not present in this file
                # If we have a parent, read it from there, otherwise keep it empty
                if self.parent:
                    sectors_read.append(self.parent.read_sectors(sector, read_count))
                else:
                    sectors_read.append(b"\x00" * read_size)
            elif bat_entry.state in (
                c_vhdx.PAYLOAD_BLOCK_UNDEFINED,
                c_vhdx.PAYLOAD_BLOCK_ZERO,
                c_vhdx.PAYLOAD_BLOCK_UNMAPPED,
            ):
                # The block is not allocated at all
                # Keep it empty
                sectors_read.append(b"\x00" * read_size)
            elif bat_entry.state == c_vhdx.PAYLOAD_BLOCK_FULLY_PRESENT:
                # The block is fully present in this file
                # Read everything from this file.
                self.fh.seek((bat_entry.file_offset_mb * MB) + (sector_in_block * self.sector_size))
                sectors_read.append(self.fh.read(read_size))
            elif bat_entry.state == c_vhdx.PAYLOAD_BLOCK_PARTIALLY_PRESENT:
                # The block is only partially present in this file
                # Read the sector bitmap to know what to read from this file and what to read from the parent
                sector_bitmap_entry = self.bat.sb(block)

                # A chunk is a group of blocks, the amount of which is determined by the chunk ratio
                # The sector bitmap spans an entire chunk, so we calculate the absolute sector number
                # in the chunk we're in.
                block_in_chunk = block % self._chunk_ratio
                sector_in_chunk = (block_in_chunk * self._sectors_per_block) + sector_in_block

                byte_idx, bit_idx = divmod(sector_in_chunk, 8)
                # Seek into the bitmap to where we are relative in the cluster
                self.fh.seek((sector_bitmap_entry.file_offset_mb * MB) + byte_idx)
                # Read the bitmap for the amount of sectors we're interested in, rounded up
                sector_bitmap = self.fh.read((read_count + 8 - 1) // 8)

                # Calculate runs from the bitmap and read from the correct source
                relative_sector = 0
                for run_type, run_count in _iter_partial_runs(sector_bitmap, bit_idx, read_count):
                    if run_type == 0:
                        # Read from parent
                        sectors_read.append(self.parent.read_sectors(sector + relative_sector, run_count))
                    else:
                        # Read from this file
                        # Here we are calculating relative to the block again
                        self.fh.seek(
                            (bat_entry.file_offset_mb * MB) + ((sector_in_block + relative_sector) * self.sector_size)
                        )
                        sectors_read.append(self.fh.read(run_count * self.sector_size))

                    relative_sector += run_count

            sector += read_count
            count -= read_count

        return b"".join(sectors_read)

    def _read(self, offset: int, length: int) -> bytes:
        sector = offset // self.sector_size
        count = (length + self.sector_size - 1) // self.sector_size

        return self.read_sectors(sector, count)


class RegionTable:
    def __init__(self, fh: BinaryIO, offset: int):
        self.fh = fh
        self.offset = offset

        fh.seek(offset)
        self.header = c_vhdx.region_table_header(fh)
        if self.header.signature != b"regi":
            raise InvalidSignature(f"Invalid region table signature: {self.header.signature}")

        self.entries = c_vhdx.region_table_entry[self.header.entry_count](fh)
        self.lookup = {UUID(bytes_le=e.guid): e for e in self.entries}

    def get(self, guid: UUID, required: bool = True) -> c_vhdx.region_table_entry | None:
        data = self.lookup.get(guid)
        if not data and required:
            raise InvalidVirtualDisk(f"Missing required region: {guid}")
        return data


class BlockAllocationTable:
    def __init__(self, vhdx: VHDX, offset: int):
        self.vhdx = vhdx
        self.offset = offset
        self.chunk_ratio = vhdx._chunk_ratio

        self._pb_count = (vhdx.size + vhdx.block_size - 1) // vhdx.block_size
        self._sb_count = (self._pb_count + self.chunk_ratio - 1) // self.chunk_ratio

        if vhdx.parent:
            self.entry_count = self._sb_count * (self.chunk_ratio + 1)
        else:
            self.entry_count = self._pb_count + ((self._pb_count - 1) // self.chunk_ratio)

        self.get = lru_cache(4096)(self.get)

    def get(self, entry: int) -> c_vhdx.bat_entry:
        """Get a BAT entry."""
        if entry + 1 > self.entry_count:
            raise ValueError(f"Invalid entry for BAT lookup: {entry} (max entry is {self.entry_count - 1})")

        self.vhdx.fh.seek(self.offset + entry * 8)
        return c_vhdx.bat_entry(self.vhdx.fh)

    def pb(self, block: int) -> c_vhdx.bat_entry:
        """Get a payload block entry for a given block."""
        # Calculate how many interleaved sector bitmap entries there must be for this block
        sb_entries = block // self.chunk_ratio
        return self.get(block + sb_entries)

    def sb(self, block: int) -> c_vhdx.bat_entry:
        """Get a sector bitmap entry for a given block."""
        # Calculate how many interleaved sector bitmap entries there must be for this block
        num_sb = block // self.chunk_ratio
        return self.get(((num_sb + 1) * self.chunk_ratio) + num_sb)


class ParentLocator:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.offset = fh.tell()
        self.header = c_vhdx.parent_locator_header(fh)
        self.type = UUID(bytes_le=self.header.locator_type)
        self._entries = c_vhdx.parent_locator_entry[self.header.key_value_count](fh)

        self.entries: dict[str, str] = {}
        for entry in self._entries:
            fh.seek(self.offset + entry.key_offset)
            key = fh.read(entry.key_length).decode("utf-16-le")
            fh.seek(self.offset + entry.value_offset)
            value = fh.read(entry.value_length).decode("utf-16-le")

            self.entries[key] = value


class MetadataTable:
    METADATA_MAP: Final[dict[UUID, Callable[[BinaryIO], Any]]] = {
        FILE_PARAMETERS_GUID: c_vhdx.file_parameters,
        VIRTUAL_DISK_SIZE_GUID: c_vhdx.virtual_disk_size,
        VIRTUAL_DISK_ID_GUID: c_vhdx.virtual_disk_id,
        LOGICAL_SECTOR_SIZE_GUID: c_vhdx.logical_sector_size,
        PHYSICAL_SECTOR_SIZE_GUID: c_vhdx.physical_sector_size,
        PARENT_LOCATOR_GUID: ParentLocator,
    }

    def __init__(self, fh: BinaryIO, offset: int, length: int):
        self.fh = fh
        self.offset = offset
        self.length = length

        fh.seek(offset)
        self.header = c_vhdx.metadata_table_header(fh)
        if self.header.signature != b"metadata":
            raise InvalidSignature(f"Invalid metadata table signature: {self.header.signature}")

        self.entries = c_vhdx.metadata_table_entry[self.header.entry_count](fh)

        self.lookup = {}
        for entry in self.entries:
            item_id = UUID(bytes_le=entry.item_id)

            fh.seek(self.offset + entry.offset)
            value = self.METADATA_MAP[item_id](fh)
            self.lookup[item_id] = value

    def get(self, guid: UUID, required: bool = True) -> Any | None:
        data = self.lookup.get(guid)
        if not data and required:
            raise InvalidVirtualDisk(f"Missing required region: {guid}")
        return data


def _iter_partial_runs(bitmap: bytes, start_idx: int, length: int) -> Iterator[tuple[int, int]]:
    current_type = (bitmap[0] & (1 << start_idx)) >> start_idx
    current_count = 0

    for byte in bitmap:
        if (current_type, byte) == (0, 0) or (current_type, byte) == (1, 0xFF):
            max_count = min(length, 8 - start_idx)
            current_count += max_count
            length -= max_count
            start_idx = 0
        else:
            for bit_idx in range(start_idx, min(length, 8)):
                sector_type = (byte & (1 << bit_idx)) >> bit_idx

                if sector_type == current_type:
                    current_count += 1
                else:
                    yield (current_type, current_count)
                    current_type = sector_type
                    current_count = 1

                length -= 1

    if current_count:
        yield (current_type, current_count)


def open_parent(path: Path, locator: dict[str, str]) -> VHDX:
    try:
        filepath = path.joinpath(locator["relative_path"].replace("\\", "/"))
        if not filepath.exists():
            filepath = path.joinpath("/" + locator["absolute_win32_path"].replace("\\", "/"))
        return VHDX(filepath)
    except Exception as e:
        raise IOError(f"Failed to open parent disk with locator {locator} from path {path}: {e}")



from __future__ import annotations


class Error(Exception):
    pass


class InvalidHeaderError(Error):
    pass


class InvalidSignature(Error):
    pass


class InvalidVirtualDisk(Error):
    pass



from __future__ import annotations

from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, TextIO
from uuid import UUID

from defusedxml import ElementTree

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element

NS = "{http://www.virtualbox.org/}"


class VBox:
    """VirtualBox XML descriptor parser.

    Args:
        fh: A file-like object of the VirtualBox XML descriptor.
    """

    def __init__(self, fh: TextIO):
        self._xml: Element = ElementTree.fromstring(fh.read())
        if self._xml.tag != f"{NS}VirtualBox":
            raise ValueError("Invalid VirtualBox XML descriptor: root element is not VirtualBox")

        if (machine := self._xml.find(f"./{NS}Machine")) is None:
            raise ValueError("Invalid VirtualBox XML descriptor: no Machine element found")

        if machine.find(f"./{NS}Hardware") is None:
            raise ValueError("Invalid VirtualBox XML descriptor: no Hardware element found")

        self.machine = Machine(self, machine)

    def __repr__(self) -> str:
        return f"<VBox uuid={self.uuid} name={self.name}>"

    @property
    def uuid(self) -> UUID | None:
        """The VM UUID."""
        return self.machine.uuid

    @property
    def name(self) -> str | None:
        """The VM name."""
        return self.machine.name

    @property
    def media(self) -> dict[UUID, HardDisk]:
        """The media (disks) registry."""
        return self.machine.media

    @property
    def hardware(self) -> Hardware:
        """The current machine hardware state."""
        return self.machine.hardware

    @property
    def snapshots(self) -> dict[UUID, Snapshot]:
        """All snapshots."""
        return self.machine.snapshots


class Machine:
    def __init__(self, vbox: VBox, element: Element):
        self.vbox = vbox
        self.element = element

    def __repr__(self) -> str:
        return f"<Machine uuid={self.uuid} name={self.name}>"

    @property
    def uuid(self) -> UUID:
        """The machine UUID."""
        return UUID(self.element.get("uuid").strip("{}"))

    @property
    def name(self) -> str:
        """The machine name."""
        return self.element.get("name")

    @property
    def current_snapshot(self) -> UUID | None:
        """The current snapshot UUID."""
        if (value := self.element.get("currentSnapshot")) is not None:
            return UUID(value.strip("{}"))
        return None

    @cached_property
    def media(self) -> dict[UUID, HardDisk]:
        """The media (disks) registry."""
        result = {}

        stack = [(None, element) for element in self.element.find(f"./{NS}MediaRegistry/{NS}HardDisks")]
        while stack:
            parent, element = stack.pop()
            hdd = HardDisk(self, element, parent)
            result[hdd.uuid] = hdd

            stack.extend([(hdd, child) for child in element.findall(f"./{NS}HardDisk")])

        return result

    @cached_property
    def hardware(self) -> Hardware:
        """The machine hardware state."""
        return Hardware(self.vbox, self.element.find(f"./{NS}Hardware"))

    @cached_property
    def snapshots(self) -> dict[UUID, Snapshot]:
        """All snapshots."""
        result = {}

        if (element := self.element.find(f"./{NS}Snapshot")) is None:
            return result

        stack = [(None, element)]
        while stack:
            parent, element = stack.pop()
            snapshot = Snapshot(self.vbox, element, parent)
            result[snapshot.uuid] = snapshot

            if (snapshots := element.find(f"./{NS}Snapshots")) is not None:
                stack.extend([(snapshot, child) for child in list(snapshots)])

        return result

    @property
    def parent(self) -> Snapshot | None:
        if (uuid := self.current_snapshot) is not None:
            return self.vbox.snapshots[uuid]
        return None


class HardDisk:
    def __init__(self, vbox: VBox, element: Element, parent: HardDisk | None = None):
        self.vbox = vbox
        self.element = element
        self.parent = parent

    def __repr__(self) -> str:
        return f"<HardDisk uuid={self.uuid} location={self.location}>"

    @property
    def uuid(self) -> UUID:
        """The disk UUID."""
        return UUID(self.element.get("uuid").strip("{}"))

    @property
    def location(self) -> str:
        """The disk location."""
        return self.element.get("location")

    @property
    def type(self) -> str | None:
        """The disk type."""
        return self.element.get("type")

    @property
    def format(self) -> str:
        """The disk format."""
        return self.element.get("format")

    @cached_property
    def properties(self) -> dict[str, str]:
        """The disk properties."""
        return {prop.get("name"): prop.get("value") for prop in self.element.findall(f"./{NS}Property")}

    @property
    def is_encrypted(self) -> bool:
        """Whether the disk is encrypted."""
        disk = self
        while disk is not None:
            if "CRYPT/KeyId" in disk.properties or "CRYPT/KeyStore" in disk.properties:
                return True
            disk = disk.parent

        return False


class Snapshot:
    def __init__(self, vbox: VBox, element: Element, parent: Snapshot | Machine | None = None):
        self.vbox = vbox
        self.element = element
        self.parent = parent

    def __repr__(self) -> str:
        return f"<Snapshot uuid={self.uuid} name={self.name}>"

    @property
    def uuid(self) -> UUID:
        """The snapshot UUID."""
        return UUID(self.element.get("uuid").strip("{}"))

    @property
    def name(self) -> str:
        """The snapshot name."""
        return self.element.get("name")

    @property
    def ts(self) -> datetime:
        """The snapshot timestamp."""
        return datetime.strptime(self.element.get("timeStamp"), "%Y-%m-%dT%H:%M:%S%z")

    @cached_property
    def hardware(self) -> Hardware:
        """The snapshot hardware state."""
        return Hardware(self.vbox, self.element.find(f"./{NS}Hardware"))


class Hardware:
    def __init__(self, vbox: VBox, element: Element):
        self.vbox = vbox
        self.element = element

    def __repr__(self) -> str:
        return f"<Hardware disks={len(self.disks)}>"

    @property
    def disks(self) -> list[HardDisk]:
        """All attached hard disks."""
        images = self.element.findall(
            f"./{NS}StorageControllers/{NS}StorageController/{NS}AttachedDevice[@type='HardDisk']/{NS}Image"
        )
        return [self.vbox.media[UUID(image.get("uuid").strip("{}"))] for image in images]
